# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""RFC1847 and RFC3156 compliant message wrapped."""
import copy
from email import message_from_string
from email.encoders import encode_7or8bit
from email.iterators import walk
from email.mime.application import MIMEApplication
from email.utils import collapse_rfc2231_value

from mailman.email.message import Message
from pgpy import PGPMessage, PGPSignature
from pgpy.constants import HashAlgorithm, SymmetricKeyAlgorithm
from public import public

from mailman_pgp.pgp.base import BaseWrapper
from mailman_pgp.utils.email import copy_headers
from mailman_pgp.utils.pgp import key_from_blob, revoc_from_blob


@public
class MIMEWrapper(BaseWrapper):
    """PGP/MIME (RFC1847 + RFC3156) compliant wrapper."""

    _signature_subtype = 'pgp-signature'
    _encryption_subtype = 'pgp-encrypted'
    _keys_subtype = 'pgp-keys'

    _signed_type = 'application/' + _signature_subtype
    _encrypted_type = 'application/' + _encryption_subtype
    _keys_type = 'application/' + _keys_subtype

    _signed_multipart = 'multipart/signed'
    _encrypted_multipart = 'multipart/encrypted'

    _signature_preamble = \
        'This is an OpenPGP/MIME signed message (RFC 4880 and 3156)'
    _encryption_preamble = \
        'This is an OpenPGP/MIME encrypted message (RFC 4880 and 3156)'

    def get_payload(self):
        yield self.msg.as_string()

    def _is_mime(self):
        is_multipart = self.msg.is_multipart()
        payloads = len(self.msg.get_payload()) if self.msg.get_payload() else 0

        return is_multipart and payloads == 2

    def is_signed(self):
        """
        Whether the whole message is MIME signed as per RFC3156 section 5.

        :return: If the message is MIME signed.
        :rtype: bool
        """
        if not self._is_mime():
            return False
        second_type = self.msg.get_payload(1).get_content_type()
        protocol_param = collapse_rfc2231_value(self.msg.get_param('protocol',
                                                                   ''))
        content_subtype = self.msg.get_content_subtype()

        return (second_type == MIMEWrapper._signed_type and
                content_subtype == 'signed' and
                protocol_param == MIMEWrapper._signed_type)

    def has_signature(self):
        return self.is_signed()

    def get_signed(self):
        """

        :return:
        """
        yield self.msg.get_payload(0).as_string()

    def get_signature(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPSignature]
        """
        try:
            sig = PGPSignature.from_blob(
                    self.msg.get_payload(1).get_payload())
        except:
            return
        yield sig

    def strip_signature(self):
        """

        :return:
        :rtype: MIMEWrapper
        """
        inner = self.msg.get_payload(0)
        copy_headers(inner, self.msg, True)
        self.msg.set_payload(inner.get_payload())
        return self

    def is_encrypted(self):
        """
        Whether the whole message is MIME encrypted as per RFC3156 section 4.

        :return: If the message is MIME encrypted.
        :rtype: bool
        """
        if not self._is_mime():
            return False
        first_part = self.msg.get_payload(0).as_string()
        first_type = self.msg.get_payload(0).get_content_type()
        second_type = self.msg.get_payload(1).get_content_type()
        content_subtype = self.msg.get_content_subtype()
        protocol_param = collapse_rfc2231_value(self.msg.get_param('protocol',
                                                                   ''))

        return ('Version: 1' in first_part and
                first_type == MIMEWrapper._encrypted_type and
                second_type == 'application/octet-stream' and
                content_subtype == 'encrypted' and
                protocol_param == MIMEWrapper._encrypted_type)

    def has_encryption(self):
        return self.is_encrypted()

    def get_encrypted(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPMessage]
        """
        try:
            msg = PGPMessage.from_blob(self.msg.get_payload(1).get_payload())
        except:
            return
        yield msg

    def is_keys(self):
        """
        Whether the message has only keys as per RFC3156 section 7.

        :return: If the message is keys.
        :rtype: bool
        """
        for part in walk(self.msg):
            if (not part.is_multipart()  # noqa
                and part.get_content_type() != MIMEWrapper._keys_type):
                return False
        return True

    def has_keys(self):
        """
        Whether the message contains keys as per RFC3156 section 7.

        :return: If the message contains keys.
        :rtype: bool
        """
        for part in walk(self.msg):
            if (not part.is_multipart()  # noqa
                and part.get_content_type() == MIMEWrapper._keys_type):
                return True
        return False

    def keys(self):
        """
        Get the collection of keys in this message.

        :return: A collection of keys.
        """
        for part in walk(self.msg):
            if (not part.is_multipart()  # noqa
                and part.get_content_type() == MIMEWrapper._keys_type):
                try:
                    key = key_from_blob(part.get_payload())
                except:
                    continue
                yield key

    def _attach_key_part(self, obj, name, description):
        key_part = MIMEApplication(_data=str(obj),
                                   _subtype=MIMEWrapper._keys_subtype,
                                   _encoder=encode_7or8bit,
                                   name=name)
        key_part.add_header('Content-Description', description)
        key_part.add_header('Content-Disposition', 'attachment',
                            filename=name)
        self.msg.attach(key_part)

    def attach_keys(self, *keys):
        """
        Attach a key to this message, as per RFC3156 section 7.

        :param keys: Keys to attach.
        :type keys: pgpy.PGPKey
        :return:
        :rtype: MIMEWrapper
        """
        if len(keys) == 0:
            return self

        if self.msg.get_content_type() != 'multipart/mixed':
            # wrap in multipart/mixed
            payload = copy.deepcopy(self.msg)
            self.msg.set_payload([])
            self.msg.set_type('multipart/mixed')
            self.msg['MIME-Version'] = '1.0'
            self.msg.attach(payload)

        for key in keys:
            filename = '0x' + key.fingerprint.keyid + '.asc'
            self._attach_key_part(key, filename, 'OpenPGP key')
        return self

    def _is_revoc(self, part):
        if part.get_content_type() != MIMEWrapper._keys_type:
            return False
        try:
            revoc_from_blob(part.get_payload())
        except ValueError:
            return False
        return True

    def is_revocs(self):
        for part in walk(self.msg):
            if (not part.is_multipart() and not self._is_revoc(part)):
                return False
        return True

    def has_revocs(self):
        for part in walk(self.msg):
            if (not part.is_multipart() and self._is_revoc(part)):
                return True
        return False

    def revocs(self):
        for part in walk(self.msg):
            if (not part.is_multipart()  # noqa
                and part.get_content_type() == MIMEWrapper._keys_type):
                try:
                    revoc = revoc_from_blob(part.get_payload())
                except:
                    continue
                yield revoc

    def attach_revocs(self, *key_revocations):
        """
        Attach a key revocation signature to the message, as a key subpart.

        :param key_revocations: A key revocation signature to attach.
        :type key_revocations: pgpy.PGPSignature
        :return:
        :rtype: MIMEWrapper
        """
        if len(key_revocations) == 0:
            return self

        if self.msg.get_content_type() != 'multipart/mixed':
            # wrap in multipart/mixed
            payload = copy.deepcopy(self.msg)
            self.msg.set_payload([])
            self.msg.set_type('multipart/mixed')
            self.msg['MIME-Version'] = '1.0'
            self.msg.attach(payload)

        for key_revocation in key_revocations:
            filename = '0x' + key_revocation.signer + '.asc'
            self._attach_key_part(key_revocation, filename,
                                  'OpenPGP key revocation')
        return self

    def verify(self, key):
        """
        Verify the signature of this message with key.

        :param key: The key to verify with.
        :type key: pgpy.PGPKey
        :return: The verified signature.
        :rtype: Generator[pgpy.types.SignatureVerification]
        """
        clear_text = next(iter(self.get_signed()))
        signature = next(iter(self.get_signature()))
        try:
            verification = key.verify(clear_text, signature)
        except:
            return
        yield verification

    def _micalg(self, hash_algo):
        algs = {
            HashAlgorithm.MD5: 'md5',
            HashAlgorithm.SHA1: 'sha1',
            HashAlgorithm.RIPEMD160: 'ripemd160',
            HashAlgorithm.SHA256: 'sha256',
            HashAlgorithm.SHA384: 'sha384',
            HashAlgorithm.SHA512: 'sha512',
            HashAlgorithm.SHA224: 'sha224'
        }
        return 'pgp-' + algs[hash_algo]

    def _wrap_signed(self, msg, signature):
        """
        As per RFC1847 and RFC3156.

        :param msg:
        :param signature:
        """
        self.msg.set_payload([])
        self.msg.attach(msg)
        self.msg.set_type(MIMEWrapper._signed_multipart)
        self.msg.set_param('micalg', self._micalg(signature.hash_algorithm))
        self.msg.set_param('protocol', MIMEWrapper._signed_type)
        self.msg.preamble = MIMEWrapper._signature_preamble

        second_part = MIMEApplication(_data=str(signature),
                                      _subtype=MIMEWrapper._signature_subtype,
                                      _encoder=encode_7or8bit,
                                      name='signature.asc')
        second_part.add_header('Content-Description',
                               'OpenPGP digital signature')
        second_part.add_header('Content-Disposition', 'attachment',
                               filename='signature.asc')
        self.msg.attach(second_part)

    def sign(self, key, **kwargs):
        """
        Sign a message with key.

        :param key: The key to sign with.
        :type key: pgpy.PGPKey
        :rtype: MIMEWrapper
        """
        payload = next(iter(self.get_payload()))
        signature = key.sign(payload, **kwargs)
        original_msg = copy.deepcopy(self.msg)
        self._wrap_signed(original_msg, signature)

        return self

    def decrypt(self, key):
        """
        Decrypt this message with key.

        :param key: The key to decrypt with.
        :type key: pgpy.PGPKey
        :return: The decrypted message.
        :rtype: mailman.email.message.Message
        """
        pmsg = next(iter(self.get_encrypted()))
        decrypted = key.decrypt(pmsg)

        dmsg = decrypted.message
        if isinstance(dmsg, bytearray):
            dmsg = dmsg.decode(decrypted.charset or 'utf-8')

        out = message_from_string(dmsg, _class=Message)
        if decrypted.is_signed:
            # rewrap, self.msg should be multipart/signed,
            # headers from out should overwrite those from self.msg
            # self.msg payload should be [out, sig]
            signature = next(iter(decrypted.signatures))
            self._wrap_signed(out, signature)
        else:
            # self.msg payload should be out.get_payload
            # headers from out should overwrite those from self.msg
            self.msg.set_payload(out.get_payload())
            copy_headers(out, self.msg, True)
        return self

    def _encrypt(self, pmsg, *keys, cipher, **kwargs):
        emsg = copy.copy(pmsg)
        if len(keys) == 1:
            emsg = keys[0].encrypt(emsg, cipher=cipher, **kwargs)
        else:
            session_key = cipher.gen_key()
            for key in keys:
                emsg = key.encrypt(emsg, cipher=cipher,
                                   sessionkey=session_key,
                                   **kwargs)
            del session_key
        return emsg

    def _wrap_encrypted(self, payload):
        self.msg.set_payload([])
        self.msg.set_type(MIMEWrapper._encrypted_multipart)
        self.msg.set_param('protocol', MIMEWrapper._encrypted_type)
        self.msg.preamble = MIMEWrapper._encryption_preamble
        first_part = MIMEApplication(_data='Version: 1',
                                     _subtype=MIMEWrapper._encryption_subtype,
                                     _encoder=encode_7or8bit)
        first_part.add_header('Content-Description',
                              'PGP/MIME version identification')
        self.msg.attach(first_part)
        second_part = MIMEApplication(_data=str(payload),
                                      _subtype='octet-stream',
                                      _encoder=encode_7or8bit,
                                      name='encrypted.asc')
        second_part.add_header('Content-Description',
                               'OpenPGP encrypted message')
        second_part.add_header('Content-Disposition', 'inline',
                               filename='encrypted.asc')
        self.msg.attach(second_part)

    def encrypt(self, *keys, cipher=SymmetricKeyAlgorithm.AES256,
                **kwargs):
        """
        Encrypt the message with key/s, using cipher.

        :param keys: The key/s to encrypt with.
        :type keys: pgpy.PGPKey
        :param cipher: The symmetric cipher to use.
        :type cipher: pgpy.constants.SymmetricKeyAlgorithm
        :return:
        :rtype: MIMEWrapper
        """
        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        if self.is_signed():
            # self.msg payload should be [ version_1, encrypted]
            # headers should remain the same, except Content-Type
            # signature should be combined into the PGP blob
            pmsg = PGPMessage.new(next(iter(self.get_signed())))
            pmsg |= next(iter(self.get_signature()))
        else:
            # self.msg payload should be [ version_1, encrypted]
            # headers should remain the same, except Content-Type
            pmsg = PGPMessage.new(next(iter(self.get_payload())))
        pmsg = self._encrypt(pmsg, *keys, cipher=cipher, **kwargs)
        self._wrap_encrypted(pmsg)
        return self

    def sign_encrypt(self, key, *keys, hash=None,
                     cipher=SymmetricKeyAlgorithm.AES256,
                     **kwargs):
        """
        Sign and encrypt the message, in one go.

        This is as per RFC 3156 section 6.2 - Combined method.

        :param key: The key to sign with.
        :type key: pgpy.PGPKey
        :param keys: The key/s to encrypt with.
        :type keys: pgpy.PGPKey
        :param hash:
        :type hash: pgpy.constants.HashAlgorithm
        :param cipher:
        :type cipher: pgpy.constants.SymmetricKeyAlgorithm
        :return:
        :rtype: MIMEWrapper
        """
        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        payload = next(iter(self.get_payload()))
        pmsg = PGPMessage.new(payload)
        pmsg |= key.sign(pmsg, hash=hash)
        pmsg = self._encrypt(pmsg, *keys, cipher=cipher, **kwargs)
        self._wrap_encrypted(pmsg)
        return self



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

import os
import unittest
from email import message_from_bytes
from operator import attrgetter

from mailman.email.message import Message
from pgpy import PGPKey
from pkg_resources import resource_string

from mailman_pgp.testing.layers import PGPLayer
from mailman_pgp.utils.pgp import revoc_from_blob


def load_blob(*path):
    return resource_string('mailman_pgp.pgp.tests',
                           os.path.join('data', *path))


def load_message(path):
    """
    :rtype: Message
    """
    return message_from_bytes(load_blob('messages', path), Message)


def load_key(path):
    """
    :rtype: pgpy.PGPKey
    """
    key, _ = PGPKey.from_blob(load_blob('keys', path))
    return key


def load_revoc(path):
    """
    :rtype: pgpy.PGPSignature
    """
    return revoc_from_blob(load_blob('revocs', path))


def payload_equal(one_msg, other_msg):
    one_payload = one_msg.get_payload()
    other_payload = other_msg.get_payload()
    if isinstance(one_payload, list) and isinstance(other_payload, list):
        if len(one_payload) != len(other_payload):
            return False
        for one_inner, other_inner in zip(one_payload, other_payload):
            if not payload_equal(one_inner, other_inner):
                return False
        return True
    else:
        return one_payload == other_payload


class WrapperTestCase(unittest.TestCase):
    layer = PGPLayer
    wrapper = None
    maxDiff = None

    def wrap(self, message):
        return self.wrapper(message)

    def is_signed(self, message, signed):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.is_signed(), signed)

    def has_signature(self, message, has):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.has_signature(), has)

    def sign(self, message, key):
        signed = self.wrap(message).sign(key)
        self.assertTrue(signed.is_signed())

    def sign_verify(self, message, priv, pub):
        signed = self.wrap(message).sign(priv)
        for signature in signed.verify(pub):
            self.assertTrue(bool(signature))

    def verify(self, message, key, valid):
        wrapped = self.wrap(message)
        for signature in wrapped.verify(key):
            self.assertEqual(bool(signature), valid)

    def is_encrypted(self, message, encrypted):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.is_encrypted(), encrypted)

    def has_encryption(self, message, has):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.has_encryption(), has)

    def encrypt(self, message, *keys, **kwargs):
        encrypted = self.wrap(message).encrypt(*keys, **kwargs)
        self.assertTrue(encrypted.is_encrypted())

    def encrypt_decrypt(self, message, pub, priv):
        decrypted = self.wrap(message).copy().encrypt(pub).decrypt(priv)

        self.assertFalse(decrypted.is_encrypted())
        self.assertTrue(payload_equal(decrypted.msg, message))

    def decrypt(self, message, key, clear):
        decrypted = self.wrap(message).copy().decrypt(key)

        self.assertFalse(decrypted.is_encrypted())
        self.assertEqual(decrypted.msg.get_payload(), clear)

    def has_keys(self, message, has_keys):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.has_keys(), has_keys)

    def is_keys(self, message, is_keys):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.is_keys(), is_keys)

    def keys(self, message, keys):
        wrapped = self.wrap(message)
        loaded = list(wrapped.keys())
        self.assertEqual(len(loaded), len(keys))

        loaded_fingerprints = list(map(attrgetter('fingerprint'), loaded))
        fingerprints = list(map(attrgetter('fingerprint'), keys))
        self.assertListEqual(loaded_fingerprints, fingerprints)

    def attach_keys(self, message, keys):
        wrapped = self.wrap(message).attach_keys(*keys)
        loaded = list(wrapped.keys())

        self.assertTrue(wrapped.has_keys())
        loaded_fingerprints = list(map(attrgetter('fingerprint'), loaded))
        fingerprints = list(map(attrgetter('fingerprint'), keys))
        self.assertListEqual(loaded_fingerprints, fingerprints)

    def has_revocs(self, message, has_revocs):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.has_revocs(), has_revocs)

    def is_revocs(self, message, is_revocs):
        wrapped = self.wrap(message)
        self.assertEqual(wrapped.is_revocs(), is_revocs)

    def revocs(self, message, revocs):
        wrapped = self.wrap(message)
        loaded = list(wrapped.revocs())
        self.assertEqual(len(loaded), len(revocs))

        loaded_issuers = list(map(attrgetter('signer'), loaded))
        issuers = list(map(attrgetter('signer'), revocs))
        self.assertListEqual(loaded_issuers, issuers)

    def attach_revocs(self, message, revocs):
        wrapped = self.wrap(message).attach_revocs(*revocs)
        loaded = list(wrapped.revocs())

        self.assertTrue(wrapped.has_revocs())
        loaded_issuers = list(map(attrgetter('signer'), loaded))
        issuers = list(map(attrgetter('signer'), revocs))
        self.assertListEqual(loaded_issuers, issuers)

    def sign_encrypt_decrypt_verify(self, message, sign_key, encrypt_key):
        wrapped = self.wrap(message)
        encrypted = wrapped.copy().sign_encrypt(sign_key, encrypt_key.pubkey)
        self.assertTrue(encrypted.is_encrypted())

        decrypted = encrypted.decrypt(encrypt_key)
        self.assertTrue(decrypted.is_signed())
        self.assertFalse(decrypted.is_encrypted())

        verification = decrypted.verify(sign_key.pubkey)
        for sig in verification:
            self.assertTrue(bool(sig))
        self.assertListEqual(list(decrypted.get_signed()),
                             list(wrapped.get_payload()))

    def sign_then_encrypt_decrypt_verify(self, message, sign_key, encrypt_key):
        wrapped = self.wrap(message)
        encrypted = wrapped.copy().sign(sign_key).encrypt(encrypt_key.pubkey)
        self.assertTrue(encrypted.is_encrypted())

        decrypted = encrypted.copy().decrypt(encrypt_key)
        self.assertTrue(decrypted.is_signed())
        self.assertFalse(decrypted.is_encrypted())

        verification = decrypted.verify(sign_key.pubkey)
        for sig in verification:
            self.assertTrue(bool(sig))
        self.assertListEqual(list(decrypted.get_signed()),
                             list(wrapped.get_payload()))

    def decrypt_verify(self, message, decrypt_key, verify_key, valid):
        wrapped = self.wrap(message)
        decrypted = wrapped.copy().decrypt(decrypt_key)

        self.assertFalse(decrypted.is_encrypted())
        self.assertTrue(decrypted.is_signed())

        verification = decrypted.verify(verify_key)
        for sig in verification:
            self.assertEqual(bool(sig), valid)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from mailman_pgp.config import mm_config
from mailman_pgp.handlers.signature_strip import SignatureStrip
from mailman_pgp.pipelines.default import PGPPostingPipeline
from mailman_pgp.testing.layers import PGPConfigLayer


class TestPGPPostingPipeline(unittest.TestCase):
    layer = PGPConfigLayer

    def test_has_pipeline(self):
        self.assertIn(PGPPostingPipeline.name, mm_config.pipelines.keys())

    def test_has_handler(self):
        pipeline = PGPPostingPipeline()
        self.assertEqual(SignatureStrip, type(next(iter(pipeline))))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the MultiSig wrapper."""
from parameterized import parameterized

from mailman_pgp.pgp.mime_multisig import MIMEMultiSigWrapper
from mailman_pgp.testing.pgp import load_key, load_message, WrapperTestCase


class MultiSigWrapperTestCase(WrapperTestCase):
    wrapper = MIMEMultiSigWrapper


class TestSigning(MultiSigWrapperTestCase):
    @parameterized.expand([
        (load_message('mime_signed.eml'),
         False),
        (load_message('mime_signed_invalid.eml'),
         False),
        (load_message('mime_multisig.eml'),
         True),
        (load_message('mime_multisig_invalid.eml'),
         True),
        (load_message('clear.eml'),
         False),
        (load_message('clear_multipart.eml'),
         False)
    ])
    def test_is_signed(self, message, signed):
        self.is_signed(message, signed)

    @parameterized.expand([
        (load_message('mime_signed.eml'),
         False),
        (load_message('mime_signed_invalid.eml'),
         False),
        (load_message('mime_multisig.eml'),
         True),
        (load_message('mime_multisig_invalid.eml'),
         True),
        (load_message('clear.eml'),
         False),
        (load_message('clear_multipart.eml'),
         False)
    ])
    def test_has_signature(self, message, has):
        self.has_signature(message, has)

    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.priv.asc')),
        (load_message('clear_multipart.eml'),
         load_key('ecc_p256.priv.asc'))
    ])
    def test_sign(self, message, key):
        self.sign(message, key)

    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('rsa_1024.pub.asc')),
        (load_message('clear_multipart.eml'),
         load_key('ecc_p256.priv.asc'),
         load_key('ecc_p256.pub.asc')),
        (load_message('mime_multisig.eml'),
         load_key('ecc_p256.priv.asc'),
         load_key('ecc_p256.pub.asc'))
    ])
    def test_sign_verify(self, message, priv, pub):
        self.sign_verify(message, priv, pub)

    @parameterized.expand([
        (load_message('mime_multisig.eml'),
         load_key('rsa_1024.pub.asc'),
         True),
        (load_message('mime_multisig_invalid.eml'),
         load_key('rsa_1024.pub.asc'),
         False)
    ])
    def test_verify(self, message, key, valid):
        self.verify(message, key, valid)


class TestCombined(MultiSigWrapperTestCase):
    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('ecc_p256.priv.asc')),
        (load_message('clear_multipart.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('ecc_p256.priv.asc'))
    ])
    def test_sign_encrypt_decrypt_verify(self, message, sign_key, encrypt_key):
        self.sign_encrypt_decrypt_verify(message, sign_key, encrypt_key)

    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('ecc_p256.priv.asc')),
        (load_message('clear_multipart.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('ecc_p256.priv.asc'))
    ])
    def test_sign_then_encrypt_decrypt_verify(self, message, sign_key,
                                              encrypt_key):
        self.sign_then_encrypt_decrypt_verify(message, sign_key, encrypt_key)

    @parameterized.expand([
        (load_message('mime_encrypted_signed.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('rsa_1024.pub.asc'),
         True)
    ])
    def test_decrypt_verify(self, message, decrypt_key, verify_key, valid):
        self.decrypt_verify(message, decrypt_key, verify_key, valid)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from mailman.interfaces.pending import IPendable
from mailman.interfaces.workflows import IWorkflow
from mailman.workflows.base import Workflow
from public import public
from zope.interface import implementer

from mailman_pgp.workflows.base import PGPMixin
from mailman_pgp.workflows.key_confirm import ConfirmPubkeyMixin
from mailman_pgp.workflows.key_set import SetPubkeyMixin
from mailman_pgp.workflows.mod_approval import ModeratorKeyRevokeApprovalMixin


class KeyRevokeBase(Workflow, PGPMixin):
    def __init__(self, mlist, pgp_address=None):
        Workflow.__init__(self)
        PGPMixin.__init__(self, mlist, pgp_address)

    @classmethod
    def pendable_class(cls):
        @implementer(IPendable)
        class Pendable(dict):
            PEND_TYPE = cls.name

        return Pendable


@public
@implementer(IWorkflow)
class KeyRevokeWorkflow(KeyRevokeBase, SetPubkeyMixin, ConfirmPubkeyMixin,
                        ModeratorKeyRevokeApprovalMixin):
    name = 'pgp-key-revoke-workflow'
    description = ''
    initial_state = 'prepare'
    save_attributes = (
        'approved',
        'address_key',
        'pubkey_key',
        'pubkey_confirmed'
    )

    def __init__(self, mlist, pgp_address=None, pubkey=None,
                 pubkey_pre_confirmed=False, pre_approved=False):
        KeyRevokeBase.__init__(self, mlist, pgp_address=pgp_address)
        SetPubkeyMixin.__init__(self, pubkey=pubkey)
        ConfirmPubkeyMixin.__init__(self, pre_confirmed=pubkey_pre_confirmed)
        ModeratorKeyRevokeApprovalMixin.__init__(self,
                                                 pre_approved=pre_approved)

    def _step_prepare(self):
        self.push('mod_approval')
        self.push('pubkey_confirmation')
        self.push('pubkey_checks')



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import copy
import unittest

from mailman.app.lifecycle import create_list
from mailman.email.message import Message, MultipartDigestMessage
from mailman.interfaces.member import MemberRole
from mailman.interfaces.subscriptions import ISubscriptionManager
from mailman.interfaces.usermanager import IUserManager
from mailman.runners.command import CommandRunner
from mailman.testing.helpers import get_queue_messages, make_testable_runner
from mailman.utilities.datetime import now
from pgpy import PGPKey, PGPUID
from pgpy.constants import (
    CompressionAlgorithm, EllipticCurveOID, HashAlgorithm, KeyFlags,
    PubKeyAlgorithm, SymmetricKeyAlgorithm)
from zope.component import getUtility

from mailman_pgp.config import mm_config
from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key
from mailman_pgp.workflows.key_change import CHANGE_CONFIRM_REQUEST
from mailman_pgp.workflows.key_confirm import CONFIRM_REQUEST
from mailman_pgp.workflows.subscription import OpenSubscriptionPolicy


def _create_plain(from_hdr, to_hdr, subject_hdr, payload):
    message = Message()
    message['From'] = from_hdr
    message['To'] = to_hdr
    message['Subject'] = subject_hdr
    message.set_payload(payload)
    return message


def _create_mixed(from_hdr, to_hdr, subject_hdr):
    message = MultipartDigestMessage()
    message['From'] = from_hdr
    message['To'] = to_hdr
    message['Subject'] = subject_hdr
    message.set_payload([])
    return message


def _run_message(message, expected_responses=None, list_id='test.example.com'):
    mm_config.switchboards['command'].enqueue(message, listid=list_id)
    make_testable_runner(CommandRunner, 'command').run()
    return get_queue_messages('virgin', expected_count=expected_responses)


class TestPreDispatch(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.mlist = create_list('test@example.com')

    def test_no_arguments(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No sub-command specified', results_msg.get_payload())

    def test_wrong_subcommand(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key wrooooooong', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Wrong sub-command specified', results_msg.get_payload())

    def test_no_pgp_list(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key set', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn("This mailing list doesn't have pgp enabled.",
                      results_msg.get_payload())


class TestPreSubscription(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.mlist = create_list('test@example.com', style_name='pgp-default')
        self.pgp_list = PGPMailingList.for_list(self.mlist)
        self.pgp_list.key = load_key('ecc_p256.priv.asc')

        self.bart_key = load_key('rsa_1024.priv.asc')
        self.anne_key = load_key('ecc_p256.priv.asc')

        self.unusable_key = PGPKey.new(PubKeyAlgorithm.ECDSA,
                                       EllipticCurveOID.SECP256K1)
        uid = PGPUID.new('Bart Person', email='bart@example.com')
        self.unusable_key.add_uid(uid,
                                  usage={KeyFlags.Certify,
                                         KeyFlags.Authentication,
                                         KeyFlags.Sign},
                                  hashes=[HashAlgorithm.SHA256,
                                          HashAlgorithm.SHA512],
                                  ciphers=[SymmetricKeyAlgorithm.AES256],
                                  compression=[CompressionAlgorithm.ZLIB]
                                  )

    def test_set(self):
        self.mlist.subscription_policy = OpenSubscriptionPolicy
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()
        token, token_owner, member = ISubscriptionManager(self.mlist).register(
                bart)

        get_queue_messages('virgin')

        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'Re: key set {}'.format(token))
        MIMEWrapper(set_message).attach_keys(self.bart_key.pubkey)

        items = _run_message(set_message, 2)

        pgp_address = PGPAddress.for_address(bart)
        self.assertIsNotNone(pgp_address)
        self.assertEqual(pgp_address.key.fingerprint,
                         self.bart_key.fingerprint)
        self.assertFalse(pgp_address.key_confirmed)

        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            results = items[0].msg
            confirm_request = items[1].msg
        else:
            results = items[1].msg
            confirm_request = items[0].msg

        self.assertIn('Key succesfully set.', results.get_payload())
        self.assertIn('Key fingerprint: {}'.format(self.bart_key.fingerprint),
                      results.get_payload())

        confirm_wrapped = PGPWrapper(confirm_request)
        self.assertTrue(confirm_wrapped.is_encrypted())

    def test_set_encrypted(self):
        self.mlist.subscription_policy = OpenSubscriptionPolicy
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()
        token, token_owner, member = ISubscriptionManager(self.mlist).register(
                bart)

        get_queue_messages('virgin')

        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'Re: key set {}'.format(token))
        MIMEWrapper(set_message).attach_keys(self.bart_key.pubkey).encrypt(
                self.pgp_list.pubkey,
                self.bart_key.pubkey)

        items = _run_message(set_message, 2)

        pgp_address = PGPAddress.for_address(bart)
        self.assertIsNotNone(pgp_address)
        self.assertEqual(pgp_address.key.fingerprint,
                         self.bart_key.fingerprint)
        self.assertFalse(pgp_address.key_confirmed)

        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            results = items[0].msg
            confirm_request = items[1].msg
        else:
            results = items[1].msg
            confirm_request = items[0].msg

        self.assertIn('Key succesfully set.', results.get_payload())
        self.assertIn('Key fingerprint: {}'.format(self.bart_key.fingerprint),
                      results.get_payload())

        confirm_wrapped = PGPWrapper(confirm_request)
        self.assertTrue(confirm_wrapped.is_encrypted())

    def test_set_no_token(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key set', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Missing token.', results_msg.get_payload())

    def test_set_no_key(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key set token', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No keys attached? Send a key.',
                      results_msg.get_payload())

    def test_set_multiple_keys(self):
        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'Re: key set token')
        MIMEWrapper(set_message).attach_keys(self.bart_key.pubkey).attach_keys(
                self.anne_key.pubkey)

        items = _run_message(set_message, 1)
        results_msg = items[0].msg

        self.assertIn('More than one key! Send only one key.',
                      results_msg.get_payload())

    def test_set_private_key(self):
        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'Re: key set token')
        MIMEWrapper(set_message).attach_keys(self.bart_key)

        items = _run_message(set_message, 1)
        results_msg = items[0].msg

        self.assertIn('You probably wanted to send your public key only.',
                      results_msg.get_payload())

    def test_set_no_encrypt_key(self):
        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'Re: key set token')
        MIMEWrapper(set_message).attach_keys(self.unusable_key.pubkey)

        items = _run_message(set_message, 1)
        results_msg = items[0].msg

        self.assertIn(
                'Need a key which can be used to encrypt communications.',
                results_msg.get_payload())

    def test_set_no_email(self):
        message = _create_mixed('', 'test@example.com', 'key set token')
        MIMEWrapper(message).attach_keys(self.bart_key.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No email to subscribe with.', results_msg.get_payload())

    def test_set_no_address(self):
        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'key set token')
        MIMEWrapper(set_message).attach_keys(self.bart_key.pubkey)

        items = _run_message(set_message, 1)
        results_msg = items[0].msg

        self.assertIn('No adddress to subscribe with.',
                      results_msg.get_payload())

    def test_set_no_pgp_address(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'key set token')
        MIMEWrapper(set_message).attach_keys(self.bart_key.pubkey)

        items = _run_message(set_message, 1)
        results_msg = items[0].msg

        self.assertIn('A pgp enabled address not found.',
                      results_msg.get_payload())

    def test_set_wrong_token(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        with transaction() as t:
            pgp_address = PGPAddress(bart)
            t.add(pgp_address)

        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'key set token')
        MIMEWrapper(set_message).attach_keys(self.bart_key.pubkey)

        items = _run_message(set_message, 1)
        results_msg = items[0].msg

        self.assertIn('Wrong token.', results_msg.get_payload())

    def test_confirm(self):
        self.mlist.subscription_policy = OpenSubscriptionPolicy
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        token, token_owner, member = ISubscriptionManager(self.mlist).register(
                bart, pubkey=self.bart_key.pubkey)

        get_queue_messages('virgin')

        message = _create_plain('bart@example.com', 'test@example.com',
                                'Re: key confirm {}'.format(token),
                                CONFIRM_REQUEST.format(
                                        self.bart_key.fingerprint,
                                        token))
        MIMEWrapper(message).sign(self.bart_key)

        _run_message(message)

        pgp_address = PGPAddress.for_address(bart)
        self.assertTrue(pgp_address.key_confirmed)
        self.assertTrue(self.mlist.is_subscribed(bart))

    def test_confirm_encrypted(self):
        self.mlist.subscription_policy = OpenSubscriptionPolicy
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        token, token_owner, member = ISubscriptionManager(self.mlist).register(
                bart, pubkey=self.bart_key.pubkey)

        get_queue_messages('virgin')

        message = _create_plain('bart@example.com', 'test@example.com',
                                'Re: key confirm {}'.format(token),
                                CONFIRM_REQUEST.format(
                                        self.bart_key.fingerprint,
                                        token))
        MIMEWrapper(message).sign_encrypt(self.bart_key,
                                          self.pgp_list.pubkey,
                                          self.bart_key.pubkey)

        _run_message(message)

        pgp_address = PGPAddress.for_address(bart)
        self.assertTrue(pgp_address.key_confirmed)
        self.assertTrue(self.mlist.is_subscribed(bart))

    def test_confirm_no_token(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key confirm', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Missing token.', results_msg.get_payload())

    def test_confirm_no_email(self):
        message = _create_plain('', 'test@example.com',
                                'key confirm token', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No email to subscribe with.', results_msg.get_payload())

    def test_confirm_no_pgp_address(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key confirm token', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('A pgp enabled address not found.',
                      results_msg.get_payload())

    def test_confirm_no_key(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            t.add(pgp_address)

        message = _create_plain('bart@example.com', 'test@example.com',
                                'Re: key confirm token',
                                CONFIRM_REQUEST.format(
                                        self.bart_key.fingerprint,
                                        'token'))
        MIMEWrapper(message).sign(self.bart_key)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No key set.', results_msg.get_payload())

    def test_confirm_not_signed(self):
        self.mlist.subscription_policy = OpenSubscriptionPolicy
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        token, token_owner, member = ISubscriptionManager(self.mlist).register(
                bart, pubkey=self.bart_key.pubkey)

        get_queue_messages('virgin')

        message = _create_plain('bart@example.com', 'test@example.com',
                                'Re: key confirm {}'.format(token),
                                CONFIRM_REQUEST.format(
                                        self.bart_key.fingerprint,
                                        token))

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Message not signed, ignoring.',
                      results_msg.get_payload())

    def test_confirm_invalid_sig(self):
        self.mlist.subscription_policy = OpenSubscriptionPolicy
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        token, token_owner, member = ISubscriptionManager(self.mlist).register(
                bart, pubkey=self.bart_key.pubkey)

        get_queue_messages('virgin')

        message = _create_plain('bart@example.com', 'test@example.com',
                                'Re: key confirm {}'.format(token),
                                CONFIRM_REQUEST.format(
                                        self.bart_key.fingerprint,
                                        token))
        MIMEWrapper(message).sign(self.bart_key)
        message.get_payload(0).set_payload(
                'Something that was definitely not signed.')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Message failed to verify.',
                      results_msg.get_payload())

    def test_confirm_wrong_token(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            t.add(pgp_address)

        message = _create_plain('bart@example.com', 'test@example.com',
                                'Re: key confirm token',
                                CONFIRM_REQUEST.format(
                                        self.bart_key.fingerprint,
                                        'token'))
        MIMEWrapper(message).sign(self.bart_key)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Wrong token.', results_msg.get_payload())

    def test_confirm_no_signed_statement(self):
        self.mlist.subscription_policy = OpenSubscriptionPolicy
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        bart.verified_on = now()

        token, token_owner, member = ISubscriptionManager(self.mlist).register(
                bart, pubkey=self.bart_key.pubkey)

        get_queue_messages('virgin')

        message = _create_plain('bart@example.com', 'test@example.com',
                                'Re: key confirm {}'.format(token),
                                'Some text, that definitely does not'
                                'contain the required/expected statement.')
        MIMEWrapper(message).sign(self.bart_key)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn("Message doesn't contain the expected statement.",
                      results_msg.get_payload())


class TestAfterSubscription(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
        with transaction():
            self.pgp_list = PGPMailingList.for_list(self.mlist)
            self.pgp_list.key = load_key('ecc_p256.priv.asc')
            self.pgp_list.key_change_workflow = 'pgp-key-change-workflow'

        self.bart_key = load_key('rsa_1024.priv.asc')
        self.bart_new_key = load_key('ecc_p256.priv.asc')

        self.unusable_key = PGPKey.new(PubKeyAlgorithm.ECDSA,
                                       EllipticCurveOID.SECP256K1)
        uid = PGPUID.new('Bart Person', email='bart@example.com')
        self.unusable_key.add_uid(uid,
                                  usage={KeyFlags.Certify,
                                         KeyFlags.Authentication,
                                         KeyFlags.Sign},
                                  hashes=[HashAlgorithm.SHA256,
                                          HashAlgorithm.SHA512],
                                  ciphers=[SymmetricKeyAlgorithm.AES256],
                                  compression=[CompressionAlgorithm.ZLIB]
                                  )

    def test_change(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')

        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')
        MIMEWrapper(message).attach_keys(self.bart_new_key.pubkey)

        items = _run_message(message, 2)
        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            results = items[0].msg
            confirm_request = items[1].msg
        else:
            results = items[1].msg
            confirm_request = items[0].msg

        self.assertIn('Key change request received.', results.get_payload())

        confirm_wrapped = PGPWrapper(confirm_request)
        self.assertTrue(confirm_wrapped.is_encrypted())
        decrypted = confirm_wrapped.decrypt(self.bart_new_key).msg
        self.assertIn('key confirm', str(decrypted['subject']))

    def test_change_encrypted(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')

        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')
        MIMEWrapper(message).attach_keys(self.bart_new_key.pubkey).encrypt(
                self.pgp_list.pubkey)

        items = _run_message(message, 2)
        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            results = items[0].msg
            confirm_request = items[1].msg
        else:
            results = items[1].msg
            confirm_request = items[0].msg

        self.assertIn('Key change request received.', results.get_payload())

        confirm_wrapped = PGPWrapper(confirm_request)
        self.assertTrue(confirm_wrapped.is_encrypted())
        decrypted = confirm_wrapped.decrypt(self.bart_new_key).msg
        self.assertIn('key confirm', str(decrypted['subject']))

    def test_change_confirm(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')

        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')
        MIMEWrapper(message).attach_keys(self.bart_new_key.pubkey)

        items = _run_message(message, 2)
        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            confirm_request = items[1].msg
        else:
            confirm_request = items[0].msg
        request_wrapped = PGPWrapper(confirm_request)
        decrypted = request_wrapped.decrypt(self.bart_new_key).msg

        subj = decrypted['subject']
        token = str(subj).split(' ')[-1]

        confirm_message = _create_plain('bart@example.com', 'test@example.com',
                                        decrypted['subject'],
                                        CHANGE_CONFIRM_REQUEST.format(
                                                self.bart_new_key.fingerprint,
                                                token))
        MIMEWrapper(confirm_message).sign(self.bart_key)

        _run_message(confirm_message)

        pgp_address = PGPAddress.for_address(bart)
        self.assertEqual(pgp_address.key_fingerprint,
                         self.bart_new_key.fingerprint)
        self.assertTrue(pgp_address.key_confirmed)

    def test_change_extra_arg(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key change extra arguments', '')
        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Extraneous argument/s: extra,arguments',
                      results_msg.get_payload())

    def test_change_no_email(self):
        message = _create_mixed('', 'test@example.com', 'key change')
        MIMEWrapper(message).attach_keys(self.bart_key.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No email to change key of.', results_msg.get_payload())

    def test_change_no_pgp_address(self):
        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')
        MIMEWrapper(message).attach_keys(self.bart_key.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('A pgp enabled address not found.',
                      results_msg.get_payload())

    def test_change_no_key_set(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn("You currently don't have a key set.",
                      results_msg.get_payload())

    def test_change_key_not_confirmed(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Your key is currently not confirmed.',
                      results_msg.get_payload())

    def test_change_no_key(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_plain('bart@example.com', 'test@example.com',
                                'key change', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No keys attached? Send a key.',
                      results_msg.get_payload())

    def test_change_multiple_keys(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        set_message = _create_mixed('bart@example.com', 'test@example.com',
                                    'key change')

        MIMEWrapper(set_message).attach_keys(self.bart_key.pubkey).attach_keys(
                self.bart_new_key.pubkey)

        items = _run_message(set_message, 1)
        results_msg = items[0].msg

        self.assertIn('More than one key! Send only one key.',
                      results_msg.get_payload())

    def test_change_private_key(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')
        MIMEWrapper(message).attach_keys(self.bart_key)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('You probably wanted to send your public key only.',
                      results_msg.get_payload())

    def test_change_no_encrypt_key(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key change')
        MIMEWrapper(message).attach_keys(self.unusable_key.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn(
                'Need a key which can be used to encrypt communications.',
                results_msg.get_payload())

    def test_revoke_resets(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        revoc = self.bart_key.revoke(self.bart_key)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key revoke')
        MIMEWrapper(message).attach_revocs(revoc)

        items = _run_message(message, 2)
        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            results_msg = items[0].msg
        else:
            results_msg = items[1].msg

        self.assertIsNone(pgp_address.key)
        self.assertFalse(pgp_address.key_confirmed)

        self.assertIn('Key needs to be reset.', results_msg.get_payload())

    def test_revoke_updates(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')

        test_key = PGPKey.new(PubKeyAlgorithm.RSAEncryptOrSign, 1024)
        uid = PGPUID.new('Some Name', email='anne@example.org')
        test_key.add_uid(uid,
                         usage={KeyFlags.Certify,
                                KeyFlags.EncryptCommunications,
                                KeyFlags.Sign},
                         hashes=[HashAlgorithm.SHA256,
                                 HashAlgorithm.SHA512],
                         ciphers=[SymmetricKeyAlgorithm.AES256],
                         compression=[CompressionAlgorithm.ZLIB])
        sub = PGPKey.new(PubKeyAlgorithm.ECDH, EllipticCurveOID.SECP256K1)
        test_key.add_subkey(sub, usage={KeyFlags.EncryptCommunications})

        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = test_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        revoc = test_key.revoke(sub.pubkey)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key revoke')
        MIMEWrapper(message).attach_revocs(revoc)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Key succesfully updated.', results_msg.get_payload())
        sub = next(iter(pgp_address.key.subkeys.values()))
        revocs = list(sub.revocation_signatures)
        self.assertEqual(len(revocs), 1)
        self.assertEqual(revoc.hash2, revocs[0].hash2)

    def test_revoke_encrypted(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        revoc = self.bart_key.revoke(self.bart_key)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key revoke')
        MIMEWrapper(message).attach_revocs(revoc).encrypt(self.pgp_list.pubkey)

        items = _run_message(message, 2)
        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            results_msg = items[0].msg
        else:
            results_msg = items[1].msg

        self.assertIsNone(pgp_address.key)
        self.assertFalse(pgp_address.key_confirmed)

        self.assertIn('Key needs to be reset.', results_msg.get_payload())

    def test_revoke_extra_arg(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key revoke extra arguments', '')
        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Extraneous argument/s: extra,arguments',
                      results_msg.get_payload())

    def test_revoke_no_email(self):
        message = _create_mixed('', 'test@example.com', 'key revoke')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No email to revoke key of.', results_msg.get_payload())

    def test_revoke_no_pgp_address(self):
        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key revoke')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('A pgp enabled address not found.',
                      results_msg.get_payload())

    def test_revoke_no_key_set(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key revoke')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn("You currently don't have a key set.",
                      results_msg.get_payload())

    def test_revoke_key_not_confirmed(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key revoke')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Your key is currently not confirmed.',
                      results_msg.get_payload())

    def test_revoke_no_revocs(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_plain('bart@example.com', 'test@example.com',
                                'key revoke', '')
        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No key revocations attached? Send a key revocation.',
                      results_msg.get_payload())

    def test_sign(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        self.mlist.subscribe(bart)
        get_queue_messages('virgin')

        self.pgp_list.key_signing_allowed = {MemberRole.member}

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')

        new_key = copy.copy(self.pgp_list.pubkey)
        uid = next(iter(new_key.userids))
        sig = self.bart_key.certify(uid)
        uid |= sig
        MIMEWrapper(message).attach_keys(new_key)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('List key updated with new signatures.',
                      results_msg.get_payload())

    def test_sign_encrypted(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        self.mlist.subscribe(bart)
        get_queue_messages('virgin')

        self.pgp_list.key_signing_allowed = {MemberRole.member}

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')

        new_key = copy.copy(self.pgp_list.pubkey)
        uid = next(iter(new_key.userids))
        sig = self.bart_key.certify(uid)
        uid |= sig
        MIMEWrapper(message).attach_keys(new_key).encrypt(self.pgp_list.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('List key updated with new signatures.',
                      results_msg.get_payload())

    def test_sign_extra_arg(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key sign extra arguments', '')
        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Extraneous argument/s: extra,arguments',
                      results_msg.get_payload())

    def test_sign_no_email(self):
        message = _create_mixed('', 'test@example.com', 'key sign')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No email.', results_msg.get_payload())

    def test_sign_no_pgp_address(self):
        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('A pgp enabled address not found.',
                      results_msg.get_payload())

    def test_sign_no_key_set(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn("You currently don't have a key set.",
                      results_msg.get_payload())

    def test_sign_key_not_confirmed(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Your key is currently not confirmed.',
                      results_msg.get_payload())

    def test_sign_no_key(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_plain('bart@example.com', 'test@example.com',
                                'key sign', '')

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No keys attached? Send a key.',
                      results_msg.get_payload())

    def test_sign_multiple_keys(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')

        MIMEWrapper(message).attach_keys(self.bart_key.pubkey).attach_keys(
                self.bart_new_key.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('More than one key! Send only one key.',
                      results_msg.get_payload())

    def test_sign_not_allowed(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        self.pgp_list.key_signing_allowed = {MemberRole.owner}

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')
        MIMEWrapper(message).attach_keys(self.pgp_list.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('You are not allowed to sign the list key.',
                      results_msg.get_payload())

    def test_sign_wrong_keymaterial(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        self.mlist.subscribe(bart)
        get_queue_messages('virgin')

        self.pgp_list.key_signing_allowed = {MemberRole.member}

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')
        MIMEWrapper(message).attach_keys(self.bart_key.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('You sent a wrong key.',
                      results_msg.get_payload())

    def test_sign_no_uids(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        self.mlist.subscribe(bart)
        get_queue_messages('virgin')

        self.pgp_list.key_signing_allowed = {MemberRole.member}

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')
        new_key = copy.copy(self.pgp_list.pubkey)
        for uid in new_key.userids:
            new_key.del_uid(uid.email)

        MIMEWrapper(message).attach_keys(new_key)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No signed UIDs found.',
                      results_msg.get_payload())

    def test_sign_no_new_sig(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        self.mlist.subscribe(bart)
        get_queue_messages('virgin')

        self.pgp_list.key_signing_allowed = {MemberRole.member}

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')
        MIMEWrapper(message).attach_keys(self.pgp_list.pubkey)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No new certifications found.',
                      results_msg.get_payload())

    def test_sign_no_sig_by_key(self):
        bart = getUtility(IUserManager).create_address('bart@example.com',
                                                       'Bart Person')
        with transaction() as t:
            pgp_address = PGPAddress(bart)
            pgp_address.key = self.bart_key.pubkey
            pgp_address.key_confirmed = True
            t.add(pgp_address)

        self.mlist.subscribe(bart)
        get_queue_messages('virgin')

        self.pgp_list.key_signing_allowed = {MemberRole.member}

        message = _create_mixed('bart@example.com', 'test@example.com',
                                'key sign')

        new_key = copy.copy(self.pgp_list.pubkey)
        uid = next(iter(new_key.userids))
        sig = self.bart_new_key.certify(uid)
        uid |= sig
        MIMEWrapper(message).attach_keys(new_key)

        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No new certifications found.',
                      results_msg.get_payload())


class TestGeneral(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.mlist = create_list('test@example.com', style_name='pgp-default')
        self.pgp_list = PGPMailingList.for_list(self.mlist)
        self.pgp_list.key = load_key('ecc_p256.priv.asc')

    def test_receive(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key receive', '')
        items = _run_message(message, 2)
        if (items[0].msg['Subject'] ==
                'The results of your email commands'):  # pragma: no cover
            pubkey_message = items[1].msg
        else:
            pubkey_message = items[0].msg

        wrapped = PGPWrapper(pubkey_message)
        self.assertTrue(wrapped.has_keys())
        keys = list(wrapped.keys())
        self.assertEqual(len(keys), 1)
        self.assertEqual(keys[0].fingerprint, self.pgp_list.key.fingerprint)

    def test_receive_extra_arg(self):
        message = _create_plain('bart@example.com', 'test@example.com',
                                'key receive extra arguments', '')
        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('Extraneous argument/s: extra,arguments',
                      results_msg.get_payload())

    def test_receive_no_email(self):
        message = _create_plain('', 'test@example.com', 'key receive', '')
        items = _run_message(message, 1)
        results_msg = items[0].msg

        self.assertIn('No email to send list public key.',
                      results_msg.get_payload())



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""The key email command."""
import copy
import logging
from email.mime.text import MIMEText

from mailman.email.message import UserNotification
from mailman.interfaces.command import ContinueProcessing, IEmailCommand
from mailman.interfaces.member import MemberRole
from mailman.interfaces.pending import IPendings
from mailman.interfaces.subscriptions import ISubscriptionManager
from mailman.interfaces.usermanager import IUserManager
from pgpy.constants import KeyFlags
from pgpy.errors import PGPError
from public import public
from zope.component import getUtility
from zope.interface import implementer

from mailman_pgp.config import mm_config
from mailman_pgp.database import transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.utils.email import get_email
from mailman_pgp.utils.pgp import key_merge, key_usable
from mailman_pgp.workflows.key_change import (CHANGE_CONFIRM_REQUEST,
                                              KeyChangeModWorkflow,
                                              KeyChangeWorkflow)
from mailman_pgp.workflows.key_confirm import CONFIRM_REQUEST
from mailman_pgp.workflows.key_revoke import KeyRevokeWorkflow

log = logging.getLogger('mailman.plugin.pgp.commands')


def _cmd_set(pgp_list, mlist, msg, msgdata, arguments, results):
    """
    `key set "token"` command.

    Used during the subscription to a PGP enabled mailing list, if the user
    didn't already setup a `PGPaddress`.

     * This command message CAN be encrypted to the list key, in which case it
     will be decrypted.
     * This command message MUST have exactly one transferrable PGP public key
     attached (either PGP/MIME or inline PGP).
    """
    if len(arguments) != 2:
        print('Missing token.', file=results)
        return ContinueProcessing.no

    wrapped = PGPWrapper(msg, True)
    if wrapped.is_encrypted():
        wrapped.try_decrypt(pgp_list.key)

    if not wrapped.has_keys():
        print('No keys attached? Send a key.', file=results)
        return ContinueProcessing.no

    keys = list(wrapped.keys())
    if len(keys) != 1:
        print('More than one key! Send only one key.', file=results)
        return ContinueProcessing.no
    key = keys.pop()

    if not key.is_public:
        print('You probably wanted to send your public key only.',
              file=results)
        return ContinueProcessing.no

    if not key_usable(key, {KeyFlags.EncryptCommunications, KeyFlags.Sign}):
        print('Need a key which can be used to encrypt communications.',
              file=results)
        return ContinueProcessing.no

    email = get_email(msg)
    if not email:
        print('No email to subscribe with.', file=results)
        return ContinueProcessing.no

    address = getUtility(IUserManager).get_address(email)
    if not address:
        print('No adddress to subscribe with.', file=results)
        return ContinueProcessing.no

    pgp_address = PGPAddress.for_address(address)
    if pgp_address is None:
        print('A pgp enabled address not found.', file=results)
        return ContinueProcessing.no

    token = arguments[1]
    pendable = getUtility(IPendings).confirm(token, expunge=False)
    if pendable is None:
        print('Wrong token.', file=results)
        return ContinueProcessing.no

    with transaction():
        pgp_address.key = key
    ISubscriptionManager(mlist).confirm(token)

    print('Key succesfully set.', file=results)
    print('Key fingerprint: {}'.format(pgp_address.key.fingerprint),
          file=results)

    return ContinueProcessing.no


def _cmd_confirm(pgp_list, mlist, msg, msgdata, arguments, results):
    """
    `key confirm "token"` command.

    Used during subscription to confirm the setting of a users key, also for
    confirming the change of a key of a subscriber after the `key change`
    command.

     * This command message CAN be encrypted to the list key, in which case it
     will be decrypted.
     * This command message MUST contain the appropriate statement of key
     ownership, sent to the user after the `key set` or `key change` commands.
     This statement MUST be singed by the users current key.
    """
    if len(arguments) != 2:
        print('Missing token.', file=results)
        return ContinueProcessing.no

    email = get_email(msg)
    if not email:
        print('No email to subscribe with.', file=results)
        return ContinueProcessing.no

    pgp_address = PGPAddress.for_email(email)
    if pgp_address is None:
        print('A pgp enabled address not found.', file=results)
        return ContinueProcessing.no

    if pgp_address.key is None:
        print('No key set.', file=results)
        return ContinueProcessing.no

    wrapped = PGPWrapper(msg, True)
    if wrapped.is_encrypted():
        wrapped.try_decrypt(pgp_list.key)

    if not wrapped.is_signed():
        print('Message not signed, ignoring.', file=results)
        return ContinueProcessing.no

    if not wrapped.verifies(pgp_address.key):
        print('Message failed to verify.', file=results)
        return ContinueProcessing.no

    token = arguments[1]

    pendable = getUtility(IPendings).confirm(token, expunge=False)
    if pendable is None:
        print('Wrong token.', file=results)
        return ContinueProcessing.no

    if pendable.get('type') in (KeyChangeWorkflow.name,
                                KeyChangeModWorkflow.name):
        expecting = CHANGE_CONFIRM_REQUEST.format(pendable.get('fingerprint'),
                                                  token)
    else:
        expecting = CONFIRM_REQUEST.format(pgp_address.key_fingerprint, token)

    for sig_subject in wrapped.get_signed():
        if expecting in sig_subject:
            with transaction():
                pgp_address.key_confirmed = True
            ISubscriptionManager(mlist).confirm(token)
            break
    else:
        print("Message doesn't contain the expected statement.", file=results)
    return ContinueProcessing.no


def _cmd_change(pgp_list, mlist, msg, msgdata, arguments, results):
    """
    `key change` command.

    Used when a user wants to change a key of a `PGPAddress`.

     * This command message CAN be encrypted to the list key, in which case it
     will be decrypted.
     * This command message MUST have exactly one transferrable PGP public key
     attached (either PGP/MIME or inline PGP).
    """
    if len(arguments) != 1:
        print('Extraneous argument/s: ' + ','.join(arguments[1:]),
              file=results)
        return ContinueProcessing.no

    email = get_email(msg)
    if not email:
        print('No email to change key of.', file=results)
        return ContinueProcessing.no

    pgp_address = PGPAddress.for_email(email)
    if pgp_address is None:
        print('A pgp enabled address not found.', file=results)
        return ContinueProcessing.no

    if pgp_address.key is None:
        print("You currently don't have a key set.", file=results)
        return ContinueProcessing.no

    if not pgp_address.key_confirmed:
        print('Your key is currently not confirmed.', file=results)
        return ContinueProcessing.no

    wrapped = PGPWrapper(msg, True)
    if wrapped.is_encrypted():
        wrapped.try_decrypt(pgp_list.key)

    if not wrapped.has_keys():
        print('No keys attached? Send a key.', file=results)
        return ContinueProcessing.no

    keys = list(wrapped.keys())
    if len(keys) != 1:
        print('More than one key! Send only one key.', file=results)
        return ContinueProcessing.no
    key = keys.pop()

    if not key.is_public:
        print('You probably wanted to send your public key only.',
              file=results)
        return ContinueProcessing.no

    if not key_usable(key, {KeyFlags.EncryptCommunications, KeyFlags.Sign}):
        print('Need a key which can be used to encrypt communications.',
              file=results)
        return ContinueProcessing.no

    workflow_class = mm_config.workflows[pgp_list.key_change_workflow]

    workflow = workflow_class(mlist, pgp_address, key)
    list(workflow)
    print('Key change request received.', file=results)
    return ContinueProcessing.no


def _cmd_revoke(pgp_list, mlist, msg, msgdata, arguments, results):
    """
    `key revoke` command.

    Used when a user has to revoke a part of a key used for a `PGPAddress`.

     * This command message CAN be encrypted to the list key, in which case it
     will be decrypted.
     * This command message MUST have at least one revocation certificate
     attached.
    """
    if len(arguments) != 1:
        print('Extraneous argument/s: ' + ','.join(arguments[1:]),
              file=results)
        return ContinueProcessing.no

    email = get_email(msg)
    if not email:
        print('No email to revoke key of.', file=results)
        return ContinueProcessing.no

    pgp_address = PGPAddress.for_email(email)
    if pgp_address is None:
        print('A pgp enabled address not found.', file=results)
        return ContinueProcessing.no

    key = pgp_address.key
    if key is None:
        print("You currently don't have a key set.", file=results)
        return ContinueProcessing.no

    if not pgp_address.key_confirmed:
        print('Your key is currently not confirmed.', file=results)
        return ContinueProcessing.no

    wrapped = PGPWrapper(msg, True)
    if wrapped.is_encrypted():
        wrapped.try_decrypt(pgp_list.key)

    if not wrapped.has_revocs():
        print('No key revocations attached? Send a key revocation.',
              file=results)
        return ContinueProcessing.no

    key_copy = copy.copy(key)

    revocs = list(wrapped.revocs())
    matches = 0
    for revoc in revocs:
        old_matches = matches
        try:
            verified = key_copy.verify(key_copy, revoc)
            if verified:
                key_copy |= revoc
                matches += 1
                continue
        except PGPError:
            pass

        for subkey in key_copy.subkeys.values():
            try:
                verified = key_copy.verify(subkey, revoc)
                if verified:
                    subkey |= revoc
                    matches += 1
                    break
            except PGPError:
                pass
        # No match?
        if matches == old_matches:
            print('Revocation found for not-found key.', file=results)

    if not key_usable(key_copy,
                      {KeyFlags.EncryptCommunications, KeyFlags.Sign}):
        # Start reset process.
        with transaction():
            pgp_address.key = None
            pgp_address.key_confirmed = False
        workflow = KeyRevokeWorkflow(mlist, pgp_address)
        list(workflow)
        print('Key needs to be reset.', file=results)
    else:
        # Just update key.
        if matches > 0:
            with transaction():
                pgp_address.key = key_copy
            print('Key succesfully updated.', file=results)
        else:
            print('Nothing to do.', file=results)
    return ContinueProcessing.yes


def _cmd_sign(pgp_list, mlist, msg, msgdata, arguments, results):
    # List public key attached, signed by the users current key.
    if len(arguments) != 1:
        print('Extraneous argument/s: ' + ','.join(arguments[1:]),
              file=results)
        return ContinueProcessing.no

    email = get_email(msg)
    if not email:
        print('No email.', file=results)
        return ContinueProcessing.no

    pgp_address = PGPAddress.for_email(email)
    if pgp_address is None:
        print('A pgp enabled address not found.', file=results)
        return ContinueProcessing.no

    key = pgp_address.key
    if key is None:
        print("You currently don't have a key set.", file=results)
        return ContinueProcessing.no

    if not pgp_address.key_confirmed:
        print('Your key is currently not confirmed.', file=results)
        return ContinueProcessing.no

    wrapped = PGPWrapper(msg, True)
    if wrapped.is_encrypted():
        wrapped.try_decrypt(pgp_list.key)

    if not wrapped.has_keys():
        print('No keys attached? Send a key.', file=results)
        return ContinueProcessing.no

    keys = list(wrapped.keys())
    if len(keys) != 1:
        print('More than one key! Send only one key.', file=results)
        return ContinueProcessing.no
    key = keys.pop()

    allowed_signers = pgp_list.key_signing_allowed
    roster_map = {
        MemberRole.member: mlist.members,
        MemberRole.owner: mlist.owners,
        MemberRole.moderator: mlist.moderators,
        MemberRole.nonmember: mlist.nonmembers
    }
    allowed = False
    for allowed_role in allowed_signers:
        allowed_roster = roster_map[allowed_role]
        if allowed_roster.get_member(email) is not None:
            allowed = True
            break

    if not allowed:
        print('You are not allowed to sign the list key.', file=results)
        return ContinueProcessing.no

    try:
        key_merge(pgp_list.key, key, pgp_address.key)
    except ValueError as e:
        print(str(e), file=results)
        return ContinueProcessing.no

    pgp_list.fs_key.save()

    print('List key updated with new signatures.', file=results)
    return ContinueProcessing.yes


def _cmd_receive(pgp_list, mlist, msg, msgdata, arguments, results):
    if len(arguments) != 1:
        print('Extraneous argument/s: ' + ','.join(arguments[1:]),
              file=results)
        return ContinueProcessing.no

    email = get_email(msg)
    if not email:
        print('No email to send list public key.', file=results)
        return ContinueProcessing.no

    out = UserNotification(email, mlist.request_address,
                           '{} public key'.format(mlist.fqdn_listname))
    out.set_type('multipart/mixed')
    out['MIME-Version'] = '1.0'
    out.attach(MIMEText('Here is the public key you requested.'))
    MIMEWrapper(out).attach_keys(pgp_list.pubkey)

    out.send(mlist)
    print('Key sent.', file=results)
    return ContinueProcessing.yes


SUBCOMMANDS = {
    'set': _cmd_set,
    'confirm': _cmd_confirm,
    'change': _cmd_change,
    'revoke': _cmd_revoke,
    'sign': _cmd_sign,
    'receive': _cmd_receive
}

ARGUMENTS = '<' + '|'.join(SUBCOMMANDS) + '>'


@public
@implementer(IEmailCommand)
class KeyCommand:
    """The `key` command."""

    name = 'key'
    argument_description = ARGUMENTS
    short_description = 'PGP key management.'
    description = """\
    A general command for PGP key management for PGP enabled mailing lists.

    `key set <token>`
      A command used to set the address public key during subscription to a
      mailing list. Is only required on the first subscription of a given
      address to a PGP enabled mailing list. This command requires the command
      message to have exactly one PGP public key attached (either PGP/MIME or
      inline). This command should be encrypted to the mailing list public key.

    `key confirm <token>`
      A command used to confirm the setting of a new public key, either during
      subscription or later after a `key change` command. Is only required on
      the first subscription of a given address to a PGP enabled mailing list.
      This command requires the command message to contain the statement sent
      from the mailing list in response to the `key set <token>` command, and
      requires this statement signed by the key that was attached to the `key
      set` command message.

    `key change`
      A command used to change the address public key.

    `key revoke`
      A command used to revoke a part of/the whole address public key, when the
      user has access to a revocation certificate (but not the signing
      capability of the key to use `key change`). A revocation certificate
      must be attached to the message like a PGP key would. The revocation
      signature is verified, and the proper revocation is performed, if the
      key is left unusable after that, you will be prompted to set and confirm
      a new one.

    `key sign`
      A command used to add your signature to the list key, if you trust it. It
      requires exactly one PGP public key attached, which should be the list
      public key, signed with your address key. It is scanned for all new
      certifications by your address key and the valid ones are imported. A
      list might be configured to allow even a non-subscriber to sign its key.

    `key receive`
      A command used to request the list public key. The list public key will
      be sent in a response.
    """

    def process(self, mlist, msg, msgdata, arguments, results):
        """See `IEmailCommand`."""
        if len(arguments) == 0:
            print('No sub-command specified,'
                  ' must be one of {}.'.format(ARGUMENTS), file=results)
            return ContinueProcessing.no

        if arguments[0] not in SUBCOMMANDS:
            print('Wrong sub-command specified,'
                  ' must be one of {}.'.format(ARGUMENTS), file=results)
            return ContinueProcessing.no

        pgp_list = PGPMailingList.for_list(mlist)
        if pgp_list is None:
            print("This mailing list doesn't have pgp enabled.", file=results)
            return ContinueProcessing.no

        command = SUBCOMMANDS[arguments[0]]
        log.debug('key {}'.format(arguments[0]))
        return command(pgp_list, mlist, msg, msgdata, arguments, results)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Common database functions and class."""
import logging
from contextlib import contextmanager

from mailman.database.transaction import transaction as mailman_transaction
from public import public
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from mailman_pgp.config import config
from mailman_pgp.model.base import Base

log = logging.getLogger('mailman.plugin.pgp.database')


@public
class Database:
    """A SQLAlchemy database."""

    def __init__(self):
        self._url = config.get_value('db', 'url')
        log.debug('Creating database at {}'.format(self._url))
        self.engine = create_engine(self._url)
        self.scoped_session = scoped_session(sessionmaker(bind=self.engine))
        Base.metadata.create_all(self.engine)
        self.session.commit()
        log.debug('Database successfully created.')

    @property
    def session(self):
        """
        Get a scoped_session.

        :return: A scoped session.
        :rtype: scoped_session
        """
        return self.scoped_session()


@public
@contextmanager
def transaction():
    """
    A transaction context manager.

    :return: A session for convenience.
    :rtype: scoped_session
    """
    try:
        yield config.db.session
    except:
        config.db.session.rollback()
        raise
    else:
        config.db.session.commit()


@public
def query(cls):
    """
    A query helper.

    :param cls: Class to query.
    :return: A query on the class.
    :rtype: sqlalchemy.orm.query.Query
    """
    return config.db.session.query(cls)


mm_transaction = mailman_transaction
public(mm_transaction=mm_transaction)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest
from copy import copy
from urllib.error import HTTPError

from mailman.app.lifecycle import create_list
from mailman.interfaces.action import Action
from mailman.interfaces.member import MemberRole
from mailman.testing.helpers import call_api
from pgpy import PGPKey

from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.testing.layers import PGPRESTLayer
from mailman_pgp.testing.pgp import load_key


class TestLists(unittest.TestCase):
    layer = PGPRESTLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')

        self.list_key = load_key('ecc_p256.priv.asc')
        with transaction():
            self.pgp_list = PGPMailingList.for_list(self.mlist)
            self.pgp_list.key = self.list_key

    def test_missing_list(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'missing.example.com')
        self.assertEqual(cm.exception.code, 404)
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'missing.example.com/key')
        self.assertEqual(cm.exception.code, 404)
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'missing.example.com/pubkey')
        self.assertEqual(cm.exception.code, 404)

    def test_all_lists(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/')
        self.assertEqual(json['total_size'], 1)
        self.assertEqual(len(json['entries']), 1)
        lists = json['entries']
        plist = lists[0]
        self.assertEqual(plist['list_id'], self.mlist.list_id)

    def test_get_list(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com')
        self.assertEqual(json['list_id'], self.mlist.list_id)
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test@example.com')
        self.assertEqual(json['list_id'], self.mlist.list_id)


class TestListConfig(unittest.TestCase):
    layer = PGPRESTLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')

        self.list_key = load_key('ecc_p256.priv.asc')
        with transaction():
            self.pgp_list = PGPMailingList.for_list(self.mlist)
            self.pgp_list.key = self.list_key

    def test_get(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com')
        cfg = dict(unsigned_msg_action='reject',
                   inline_pgp_action='defer',
                   expired_sig_action='reject',
                   revoked_sig_action='reject',
                   invalid_sig_action='reject',
                   duplicate_sig_action='reject',
                   strip_original_sig=False,
                   sign_outgoing=False,
                   nonencrypted_msg_action='reject',
                   encrypt_outgoing=True,
                   key_change_workflow='pgp-key-change-mod-workflow',
                   key_signing_allowed=['moderator', 'owner'])

        for key in cfg:
            value = json[key]
            if isinstance(value, list):
                value = sorted(value)
            self.assertEqual(cfg[key], value)

    def test_put(self):
        cfg = dict(unsigned_msg_action='defer',
                   inline_pgp_action='defer',
                   expired_sig_action='defer',
                   revoked_sig_action='defer',
                   invalid_sig_action='defer',
                   duplicate_sig_action='defer',
                   strip_original_sig=False,
                   sign_outgoing=True,
                   nonencrypted_msg_action='defer',
                   encrypt_outgoing=False,
                   key_change_workflow='pgp-key-change-mod-workflow',
                   key_signing_allowed=['member', 'owner'])
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com',
                data=cfg,
                method='PUT')

        self.assertEqual(response.status_code, 204)
        for key in cfg:
            attr = getattr(self.pgp_list, key)
            if isinstance(attr, Action):
                attr = attr.name
            elif key == 'key_signing_allowed':
                attr = sorted(enum.name for enum in attr)
            self.assertEqual(attr, cfg[key])

    def test_put_wrong_value(self):
        cfg = dict(unsigned_msg_action='not-an-action',
                   inline_pgp_action='defer',
                   expired_sig_action='defer',
                   revoked_sig_action='defer',
                   invalid_sig_action='defer',
                   duplicate_sig_action='defer',
                   strip_original_sig=False,
                   sign_outgoing=True,
                   nonencrypted_msg_action='defer',
                   encrypt_outgoing=False,
                   key_change_workflow='pgp-key-change-mod-workflow',
                   key_signing_allowed=['member', 'owner'])
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'test.example.com',
                     data=cfg,
                     method='PUT')
        self.assertEqual(cm.exception.code, 400)

    def test_patch(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com',
                data=dict(unsigned_msg_action='defer'),
                method='PATCH')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(self.pgp_list.unsigned_msg_action, Action.defer)

    def test_patch_set(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com',
                data=dict(key_signing_allowed=['member', 'owner']),
                method='PATCH')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(self.pgp_list.key_signing_allowed,
                         {MemberRole.member, MemberRole.owner})

    def test_patch_wrong_attribute(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'test.example.com',
                     data=dict(something_something='somewhere'),
                     method='PATCH')
        self.assertEqual(cm.exception.code, 400)

    def test_patch_wrong_value(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'test.example.com',
                     data=dict(unsigned_msg_action='not-an-action'),
                     method='PATCH')
        self.assertEqual(cm.exception.code, 400)

    def test_missing_list(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'missing.example.com',
                     data=dict(),
                     method='PUT')
        self.assertEqual(cm.exception.code, 404)
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'missing.example.com',
                     data=dict(),
                     method='PATCH')
        self.assertEqual(cm.exception.code, 404)


class TestListKey(unittest.TestCase):
    layer = PGPRESTLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')

        self.list_key = load_key('ecc_p256.priv.asc')
        with transaction():
            self.pgp_list = PGPMailingList.for_list(self.mlist)
            self.pgp_list.key = self.list_key

    def test_get_list_key(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com/key')
        json.pop('http_etag')
        self.assertEqual(len(json.keys()), 2)
        self.assertIn('key', json.keys())
        self.assertIn('key_fingerprint', json.keys())

        key, _ = PGPKey.from_blob(json['key'])
        self.assertFalse(key.is_public)
        self.assertEqual(json['key_fingerprint'], key.fingerprint)
        self.assertEqual(self.list_key.fingerprint, key.fingerprint)

    def test_missing_list_key(self):
        with transaction():
            self.pgp_list.key = None
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'test.example.com/key')
        self.assertEqual(cm.exception.code, 404)

    def test_set_list_key(self):
        new_key = load_key('rsa_1024.priv.asc')
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com/key',
                data=dict(key=str(new_key)),
                method='PUT')

        self.assertEqual(response.status_code, 202)

        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com/key')

        key, _ = PGPKey.from_blob(json['key'])
        self.assertEqual(key.fingerprint, new_key.fingerprint)

    def test_set_list_key_wrong(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'test.example.com/key',
                     dict(key='some stuff?'),
                     method='PUT')
        self.assertEqual(cm.exception.code, 400)

    def test_set_list_key_public(self):
        new_key = load_key('rsa_1024.priv.asc')
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'test.example.com/key',
                     data=dict(key=str(new_key.pubkey)),
                     method='PUT')

        self.assertEqual(cm.exception.code, 400)

        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com/key')

        key, _ = PGPKey.from_blob(json['key'])
        self.assertEqual(key.fingerprint, self.list_key.fingerprint)

    def test_get_list_pubkey(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com/pubkey')

        json.pop('http_etag')
        self.assertEqual(len(json.keys()), 2)
        self.assertIn('public_key', json.keys())
        self.assertIn('key_fingerprint', json.keys())

        key, _ = PGPKey.from_blob(json['public_key'])
        self.assertTrue(key.is_public)
        self.assertEqual(json['key_fingerprint'], key.fingerprint)
        self.assertEqual(self.list_key.fingerprint, key.fingerprint)

    def test_missing_list_pubkey(self):
        with transaction():
            self.pgp_list.key = None
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/plugins/pgp/lists/'
                     'test.example.com/pubkey')
        self.assertEqual(cm.exception.code, 404)

    def test_set_list_pubkey(self):
        signer_key = load_key('rsa_1024.priv.asc')
        signed = copy(self.list_key.pubkey)
        uid = next(iter(signed.userids))
        sig = signer_key.certify(uid)
        uid |= sig

        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com/pubkey',
                data=dict(public_key=str(signed)),
                method='PUT')

        self.assertEqual(response.status_code, 202)

        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/lists/'
                'test.example.com/pubkey')

        key, _ = PGPKey.from_blob(json['public_key'])
        uid = next(iter(key.userids))
        self.assertIn(sig, list(uid.signatures))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest

from mailman.testing.helpers import call_api

from mailman_pgp.config import config
from mailman_pgp.testing.layers import PGPRESTLayer


class TestConfig(unittest.TestCase):
    layer = PGPRESTLayer

    def setUp(self):
        pass

    def test_get_config(self):
        json, response = call_api(
                'http://localhost:9001/3.1/plugins/pgp/config/')
        for section in config.sections():
            self.assertIn(section, json)
            json_section = json[section]
            for option in config.options(section):
                self.assertIn(option, json_section)
                self.assertEqual(config.get(section, option),
                                 json_section[option])

    def test_get_section(self):
        for section in config.sections():
            json, response = call_api(
                    'http://localhost:9001/3.1/plugins/pgp/config/' +
                    section + '/')
            for option in config.options(section):
                self.assertIn(option, json)
                self.assertEqual(config.get(section, option),
                                 json[option])

    def test_get_option(self):
        for section in config.sections():
            for option in config.options(section):
                json, response = call_api(
                        'http://localhost:9001/3.1/plugins/pgp/config/' +
                        section + '/' + option + '/')
                self.assertIn(option, json)
                self.assertEqual(config.get(section, option), json[option])



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.mailinglist import Personalization
from mailman.interfaces.mta import SomeRecipientsFailed
from mailman.testing.helpers import (specialized_message_from_string as mfs,
                                     subscribe)

from mailman_pgp.database import transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.mta.deliver import deliver
from mailman_pgp.testing.layers import PGPSMTPLayer
from mailman_pgp.testing.pgp import load_key


class TestDeliver(unittest.TestCase):
    layer = PGPSMTPLayer

    def setUp(self):
        with transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
            self.mlist.personalize = Personalization.individual

        self.list_key = load_key('ecc_p256.priv.asc')
        self.pgp_list = PGPMailingList.for_list(self.mlist)
        self.pgp_list.key = self.list_key

        # Make Anne a member of this mailing list.
        self.anne = subscribe(self.mlist, 'Anne', email='anne@example.org')
        self.anne_key = load_key('rsa_1024.priv.asc')

        self.bart = subscribe(self.mlist, 'Bart', email='bart@example.org')
        self.bart_key = load_key('ecc_secp256k1.priv.asc')

        with transaction() as t:
            self.pgp_anne = PGPAddress(self.anne.address)
            self.pgp_anne.key = self.anne_key.pubkey
            self.pgp_anne.key_confirmed = True
            t.add(self.pgp_anne)

        with transaction() as t:
            self.pgp_bart = PGPAddress(self.bart.address)
            self.pgp_bart.key = self.bart_key.pubkey
            self.pgp_bart.key_confirmed = True
            t.add(self.pgp_bart)
        self.msg = mfs("""\
From: anne@example.org
To: test@example.com
Subject: some subject

Some text.
""")

    def test_deliver(self):
        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'],
                       pgp_is_posting=True)
        deliver(self.mlist, self.msg, msgdata)

    def test_deliver_no_key(self):
        with transaction():
            self.pgp_anne.key = None
        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'],
                       pgp_is_posting=True)
        with self.assertRaises(SomeRecipientsFailed) as err:
            deliver(self.mlist, self.msg, msgdata)
        self.assertEqual(err.exception.temporary_failures,
                         ['anne@example.org'])



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the combined wrapper."""
from parameterized import parameterized

from mailman_pgp.pgp.inline import InlineWrapper
from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.pgp.mime_multisig import MIMEMultiSigWrapper
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.testing.pgp import (
    load_key, load_message, load_revoc, WrapperTestCase)


class PGPWrapperTestCase(WrapperTestCase):
    wrapper = PGPWrapper


class TestSigning(PGPWrapperTestCase):
    @parameterized.expand([
        (load_message('inline_cleartext_signed.eml'),
         True),
        (load_message('inline_cleartext_signed_invalid.eml'),
         True),
        (load_message('inline_encrypted.eml'),
         False),
        (load_message('inline_signed.eml'),
         True),
        (load_message('inline_signed_corrupted.eml'),
         False),
        (load_message('mime_signed.eml'),
         True),
        (load_message('mime_signed_invalid.eml'),
         True),
        (load_message('clear.eml'),
         False)
    ])
    def test_is_signed(self, message, signed):
        self.is_signed(message, signed)

    @parameterized.expand([
        (load_message('inline_signed_multipart.eml'),
         True),
        (load_message('inline_signed.eml'),
         True),
        (load_message('mime_signed.eml'),
         True),
        (load_message('mime_signed_invalid.eml'),
         True),
        (load_message('clear.eml'),
         False)
    ])
    def test_has_signature(self, message, has):
        self.has_signature(message, has)

    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.priv.asc')),
        (load_message('clear.eml'),
         load_key('ecc_p256.priv.asc'))
    ])
    def test_sign(self, message, key):
        self.sign(message, key)

    @parameterized.expand([
        (load_message('inline_cleartext_signed.eml'),
         load_key('rsa_1024.pub.asc'),
         True),
        (load_message('inline_cleartext_signed_invalid.eml'),
         load_key('rsa_1024.pub.asc'),
         False),
        (load_message('mime_signed.eml'),
         load_key('rsa_1024.pub.asc'),
         True),
        (load_message('mime_signed_invalid.eml'),
         load_key('rsa_1024.pub.asc'),
         False)
    ])
    def test_verify(self, message, key, valid):
        self.verify(message, key, valid)


class TestEncryption(PGPWrapperTestCase):
    @parameterized.expand([
        (load_message('inline_encrypted.eml'),
         True),
        (load_message('mime_signed_then_encrypted.eml'),
         True),
        (load_message('inline_cleartext_signed.eml'),
         False),
        (load_message('inline_cleartext_signed_invalid.eml'),
         False),
        (load_message('inline_signed.eml'),
         False),
        (load_message('inline_signed_corrupted.eml'),
         False),
        (load_message('clear.eml'),
         False)
    ])
    def test_is_encrypted(self, message, encrypted):
        self.is_encrypted(message, encrypted)

    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.pub.asc')),
        (load_message('clear.eml'),
         (load_key('rsa_1024.pub.asc'),
          load_key('ecc_p256.pub.asc')))
    ])
    def test_encrypt(self, message, keys, **kwargs):
        if isinstance(keys, tuple):
            self.encrypt(message, *keys, **kwargs)
        else:
            self.encrypt(message, keys, **kwargs)

    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.pub.asc'),
         load_key('rsa_1024.priv.asc')),
        (load_message('clear.eml'),
         load_key('ecc_p256.pub.asc'),
         load_key('ecc_p256.priv.asc'))
    ])
    def test_encrypt_decrypt(self, message, pub, priv):
        self.encrypt_decrypt(message, pub, priv)

    @parameterized.expand([
        (load_message('inline_encrypted.eml'),
         load_key('rsa_1024.priv.asc'),
         'Some encrypted text.\n\n')
    ])
    def test_decrypt(self, message, key, clear):
        self.decrypt(message, key, clear)


class TestKeys(PGPWrapperTestCase):
    @parameterized.expand([
        (load_message('inline_privkey.eml'),
         True),
        (load_message('inline_pubkey.eml'),
         True),
        (load_message('inline_cleartext_signed.eml'),
         False),
        (load_message('mime_privkey.eml'),
         True),
        (load_message('mime_pubkey.eml'),
         True),
        (load_message('mime_signed.eml'),
         False),
        (load_message('clear.eml'),
         False)
    ])
    def test_has_keys(self, message, has_keys):
        self.has_keys(message, has_keys)

    @parameterized.expand([
        (load_message('inline_privkey.eml'),
         [load_key('rsa_1024.priv.asc')]),
        (load_message('inline_pubkey.eml'),
         [load_key('rsa_1024.pub.asc')]),
        (load_message('mime_privkey.eml'),
         [load_key('rsa_1024.priv.asc')]),
        (load_message('mime_pubkey.eml'),
         [load_key('rsa_1024.pub.asc')])
    ])
    def test_keys(self, message, keys):
        self.keys(message, keys)


class TestRevocs(PGPWrapperTestCase):
    @parameterized.expand([
        (load_message('mime_revoc.eml'),
         True),
        (load_message('inline_revoc.eml'),
         True),
        (load_message('inline_revoc_multipart.eml'),
         True)
    ])
    def test_has_revocs(self, message, has_revocs):
        self.has_revocs(message, has_revocs)

    @parameterized.expand([
        (load_message('mime_revoc.eml'),
         False),
        (load_message('inline_revoc.eml'),
         True),
        (load_message('inline_revoc_multipart.eml'),
         False)
    ])
    def test_is_revocs(self, message, is_revocs):
        self.is_revocs(message, is_revocs)

    @parameterized.expand([
        (load_message('mime_revoc.eml'),
         (load_revoc('rsa_1024.revoc.asc'),)),
        (load_message('inline_revoc.eml'),
         (load_revoc('rsa_1024.revoc.asc'),)),
        (load_message('inline_revoc_multipart.eml'),
         (load_revoc('rsa_1024.revoc.asc'),))
    ])
    def test_revocs(self, message, revocs):
        self.revocs(message, revocs)


class TestCombined(PGPWrapperTestCase):
    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('ecc_p256.priv.asc'))
    ])
    def test_sign_encrypt_decrypt_verify(self, message, sign_key, encrypt_key):
        self.sign_encrypt_decrypt_verify(message, sign_key, encrypt_key)

    @parameterized.expand([
        (load_message('clear.eml'),
         load_key('rsa_1024.priv.asc'),
         load_key('ecc_p256.priv.asc'))
    ])
    def test_sign_then_encrypt_decrypt_verify(self, message, sign_key,
                                              encrypt_key):
        self.sign_then_encrypt_decrypt_verify(message, sign_key, encrypt_key)


class TestPGPWrapper(PGPWrapperTestCase):
    def setUp(self):
        self.msg = load_message('clear.eml')

    def test_defaults(self):
        PGPWrapper(self.msg)
        for wrap_class in (InlineWrapper, MIMEWrapper, MIMEMultiSigWrapper):
            PGPWrapper(self.msg, default=wrap_class)
        self.assertRaises(ValueError, PGPWrapper, self.msg,
                          default='Not a class')



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest

from mailman.app.lifecycle import create_list
from mailman.testing.helpers import specialized_message_from_string as mfs

from mailman_pgp.config import mm_config
from mailman_pgp.database import mm_transaction
from mailman_pgp.rules.encryption import Encryption
from mailman_pgp.testing.layers import PGPConfigLayer


class TestPGPEncryptionRule(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.rule = Encryption()
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')

    def test_has_rule(self):
        self.assertIn(Encryption.name, mm_config.rules.keys())

    def test_matches(self):
        msgdata = {'pgp_moderate': True}
        msg = mfs("""\
From: anne@example.com
To: test@example.com

""")
        self.assertTrue(self.rule.check(self.mlist, msg, msgdata))
        msgdata['pgp_moderate'] = False
        self.assertFalse(self.rule.check(self.mlist, msg, msgdata))
        self.assertFalse(self.rule.check(self.mlist, msg, {}))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import time
import unittest
from os.path import exists, join
from tempfile import TemporaryDirectory

from mailman_pgp.model.fs_key import FSKey
from mailman_pgp.testing.layers import PGPLayer
from mailman_pgp.testing.pgp import load_key


class TestFSKey(unittest.TestCase):
    layer = PGPLayer

    def setUp(self):
        self.tmpdir = TemporaryDirectory()

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_load(self):
        key_name = 'something.asc'
        key_path = join(self.tmpdir.name, key_name)
        key_data = load_key('rsa_1024.priv.asc')
        with open(key_path, 'w') as key_file:
            key_file.write(str(key_data))

        key = FSKey(self.tmpdir.name, key_name, False)
        self.assertEqual(key.key_path, key_path)

        key.load()
        self.assertEqual(key.key.fingerprint, key_data.fingerprint)

    def test_reload(self):
        key_name = 'something.asc'
        key_path = join(self.tmpdir.name, key_name)
        key_data = load_key('rsa_1024.priv.asc')
        new_key_data = load_key('ecc_p256.priv.asc')
        with open(key_path, 'w') as key_file:
            key_file.write(str(key_data))

        key = FSKey(self.tmpdir.name, key_name, True)

        time.sleep(2)
        with open(key_path, 'w') as key_file:
            key_file.write(str(new_key_data))

        key.reload()
        self.assertIsNotNone(key.key)
        self.assertEqual(key.key.fingerprint, new_key_data.fingerprint)

    def test_reload_none(self):
        key_name = 'something.asc'
        key = FSKey(self.tmpdir.name, key_name, False)
        key_data = load_key('rsa_1024.priv.asc')
        with open(key.key_path, 'w') as key_file:
            key_file.write(str(key_data))

        self.assertIsNone(key.key)
        key.reload()
        self.assertIsNotNone(key.key)
        self.assertEqual(key.key.fingerprint, key_data.fingerprint)

    def test_save(self):
        key_name = 'something.asc'
        key = FSKey(self.tmpdir.name, key_name)
        key_data = load_key('rsa_1024.priv.asc')

        key.key = key_data
        key.save()
        self.assertTrue(exists(key.key_path))

    def test_delete(self):
        key_name = 'something.asc'
        key_path = join(self.tmpdir.name, key_name)
        key_data = load_key('rsa_1024.priv.asc')
        with open(key_path, 'w') as key_file:
            key_file.write(str(key_data))

        key = FSKey(self.tmpdir.name, key_name, True)

        key.delete()
        self.assertFalse(exists(key.key_path))
        self.assertIsNotNone(key.key)

    def test_delete_none(self):
        key = FSKey(self.tmpdir.name, 'something.asc')
        key.delete()

    def test_shred(self):
        key_name = 'something.asc'
        key_path = join(self.tmpdir.name, key_name)
        key_data = load_key('rsa_1024.priv.asc')
        with open(key_path, 'w') as key_file:
            key_file.write(str(key_data))

        key = FSKey(self.tmpdir.name, key_name, True)

        key.shred()
        self.assertFalse(exists(key.key_path))
        self.assertIsNotNone(key.key)

    def test_shred_none(self):
        key = FSKey(self.tmpdir.name, 'something.asc')
        key.shred()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from mailman.email.message import UserNotification
from mailman.interfaces.pending import IPendable, IPendings
from mailman.interfaces.subscriptions import TokenOwner
from mailman.interfaces.workflows import IWorkflow
from mailman.workflows.base import Workflow
from pgpy import PGPKey
from public import public
from zope.component import getUtility
from zope.interface import implementer

from mailman_pgp.config import config
from mailman_pgp.database import transaction
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.workflows.base import PGPMixin
from mailman_pgp.workflows.mod_approval import (
    ModeratorKeyChangeApprovalMixin)

CHANGE_CONFIRM_REQUEST = """\
----------
TODO: this is a pgp enabled list.
You requested to change your key.
Reply to this message with this whole text
signed with your supplied key, either inline or PGP/MIME.

Fingerprint: {}
Token: {}
----------
"""


class KeyChangeBase(Workflow, PGPMixin):
    save_attributes = (
        'address_key',
        'pubkey_key',
    )

    def __init__(self, mlist, pgp_address=None, pubkey=None):
        Workflow.__init__(self)
        PGPMixin.__init__(self, mlist, pgp_address)
        self.pubkey = pubkey

    @property
    def pubkey_key(self):
        return str(self.pubkey)

    @pubkey_key.setter
    def pubkey_key(self, value):
        self.pubkey, _ = PGPKey.from_blob(value)

    def _pend(self, token_owner, lifetime=None):
        pendings = getUtility(IPendings)
        pendable = self.pendable_class()(
                email=self.pgp_address.email,
                pubkey=str(self.pubkey),
                fingerprint=self.pubkey.fingerprint
        )

        self.token = pendings.add(pendable, lifetime=lifetime)
        self.token_owner = token_owner

    def _step_change_key(self):
        if self.pgp_address is None or self.pubkey is None:
            raise ValueError

        self.push('send_key_confirm_request')

    def _step_send_key_confirm_request(self):
        self._pend(TokenOwner.subscriber,
                   lifetime=config.get_value('misc',
                                             'change_request_lifetime'))
        self.push('receive_confirmation')
        self.save()
        request_address = self.mlist.request_address
        email_address = self.pgp_address.email
        msg = UserNotification(email_address, request_address,
                               'key confirm {}'.format(self.token),
                               CHANGE_CONFIRM_REQUEST.format(
                                       self.pubkey.fingerprint,
                                       self.token))
        PGPWrapper(msg).sign_encrypt(self.pgp_list.key, self.pubkey)

        msg.send(self.mlist)
        raise StopIteration

    def _step_receive_confirmation(self):
        self._set_token(TokenOwner.no_one)

    def _step_do_change(self):
        with transaction():
            self.pgp_address.key = self.pubkey
            self.pgp_address.key_confirmed = True

    @classmethod
    def pendable_class(cls):
        @implementer(IPendable)
        class Pendable(dict):
            PEND_TYPE = cls.name

        return Pendable


@public
@implementer(IWorkflow)
class KeyChangeWorkflow(KeyChangeBase):
    name = 'pgp-key-change-workflow'
    description = ''
    initial_state = 'prepare'

    def _step_prepare(self):
        self.push('do_change')
        self.push('change_key')


@public
@implementer(IWorkflow)
class KeyChangeModWorkflow(KeyChangeBase, ModeratorKeyChangeApprovalMixin):
    name = 'pgp-key-change-mod-workflow'
    description = ''
    initial_state = 'prepare'
    save_attributes = (
        'approved',
        'address_key',
        'pubkey_key'
    )

    def __init__(self, mlist, pgp_address=None, pubkey=None,
                 pre_approved=False):
        KeyChangeBase.__init__(self, mlist, pgp_address, pubkey)
        ModeratorKeyChangeApprovalMixin.__init__(self, pre_approved)

    def _step_prepare(self):
        self.push('do_change')
        self.push('mod_approval')
        self.push('change_key')



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from mailman.app.lifecycle import create_list

from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.styles.base import PGPStyle
from mailman_pgp.testing.config import patch_config
from mailman_pgp.testing.layers import PGPConfigLayer


class TestBaseStyle(unittest.TestCase):
    layer = PGPConfigLayer

    def test_apply(self):
        # Create with default style.
        mlist = create_list('test@example.com')
        # Manually apply base PGPStyle.
        base_style = PGPStyle()
        base_style.apply(mlist)

        pgp_list = PGPMailingList.for_list(mlist)

        # Test that we have our PGPMailingList
        self.assertIsNotNone(pgp_list)
        self.assertEqual(pgp_list.mlist, mlist)
        self.assertEqual(mlist.posting_chain, 'pgp-posting-chain')

        # Test another apply doesn't fail
        base_style.apply(mlist)

    @patch_config('keypairs', 'autogenerate', 'no')
    def test_autogenerate(self):
        # Create with default style.
        mlist = create_list('test@example.com')
        # Manually apply base PGPStyle.
        base_style = PGPStyle()
        base_style.apply(mlist)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import contextlib

from mailman.testing.layers import (ConfigLayer, RESTLayer, SMTPLayer)
from sqlalchemy.exc import SQLAlchemyError

from mailman_pgp.config import config
from mailman_pgp.database import transaction
from mailman_pgp.model.base import Base


def reset_rollback():
    config.db.session.rollback()


def reset_pgp_dirs():
    for keydir in (config.pgp.keydirs.values()):
        for path in keydir.iterdir():
            full_path = keydir.joinpath(path)
            if full_path.is_file():
                full_path.unlink()


def reset_pgp_hard():
    reset_rollback()
    reset_pgp_dirs()
    with transaction():
        Base.metadata.drop_all(config.db.engine)
        Base.metadata.create_all(config.db.engine)


def reset_pgp_soft():
    reset_rollback()
    reset_pgp_dirs()
    with contextlib.closing(config.db.engine.connect()) as con:
        trans = con.begin()
        for table in reversed(Base.metadata.sorted_tables):
            try:
                con.execute(table.delete())
            except SQLAlchemyError:
                pass
        trans.commit()


class PGPLayer:
    pass


# It's weird that ws have to do this, but for some reason nose2 test layers
# don't work when ws create a mixin class with the two classmethods
# and subclass both it and the respective Mailman Core test layer.
class PGPConfigLayer(ConfigLayer):
    @classmethod
    def tearDown(cls):
        reset_pgp_soft()

    @classmethod
    def testTearDown(cls):
        reset_pgp_soft()


class PGPMigrationLayer(ConfigLayer):
    @classmethod
    def tearDown(cls):
        reset_pgp_hard()

    @classmethod
    def testTearDown(cls):
        reset_pgp_hard()


class PGPSMTPLayer(SMTPLayer):
    @classmethod
    def tearDown(cls):
        reset_pgp_soft()

    @classmethod
    def testTearDown(cls):
        reset_pgp_soft()


class PGPRESTLayer(RESTLayer):
    @classmethod
    def tearDown(cls):
        reset_pgp_soft()

    @classmethod
    def testTearDown(cls):
        reset_pgp_soft()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""MIMEWrapper with multiple signature as per draft-ietf-openpgp-multsig-02."""
import copy
from email import message_from_string
from email.encoders import encode_7or8bit
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.utils import collapse_rfc2231_value

from mailman.email.message import Message
from pgpy import PGPSignature

from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.utils.email import copy_headers


class MIMEMultiSigWrapper(MIMEWrapper):
    """https://tools.ietf.org/html/draft-ietf-openpgp-multsig-02"""

    _signature_preamble = \
        'This is an OpepPGP/MIME signed message' \
        '(RFC 4880, 3156 and draft-ietf-openpgp-multsig).\n' \
        'see https://tools.ietf.org/html/draft-ietf-openpgp-multsig-02' \
        'for more details.'

    def is_signed(self):
        """
        Whether the message is signed as per draft-ietf-openpgp-multsig-02.

        :return: If the message is MIME signed.
        :rtype: bool
        """
        if not self._is_mime():
            return False
        second_part = self.msg.get_payload(1)
        second_type = second_part.get_content_type()
        protocol_param = collapse_rfc2231_value(self.msg.get_param('protocol',
                                                                   ''))
        content_subtype = self.msg.get_content_subtype()

        return (second_part.is_multipart() and
                second_type == 'multipart/mixed' and
                content_subtype == 'signed' and
                protocol_param == 'multipart/mixed' and
                all(part.get_content_type() == MIMEWrapper._signed_type
                    for part in second_part.get_payload()))

    def get_signature(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPSignature]
        """
        for part in self.msg.get_payload(1).get_payload():
            try:
                sig = PGPSignature.from_blob(part.get_payload())
            except:
                continue
            yield sig

    def strip_signature(self):
        pass

    def sign(self, key, **kwargs):
        """
        Sign a message with key.

        :param key: The key to sign with.
        :type key: pgpy.PGPKey
        :return:
        :rtype: MIMEMultiSigWrapper
        """

        if self.is_signed():
            signed = next(iter(self.get_signed()))
            signature = key.sign(signed, **kwargs)
            sig_part = MIMEApplication(_data=str(signature),
                                       _subtype=MIMEWrapper._signature_subtype,
                                       _encoder=encode_7or8bit,
                                       name='signature.asc')
            sig_part.add_header('Content-Description',
                                'OpenPGP digital signature')
            sig_part.add_header('Content-Disposition', 'attachment',
                                filename='signature.asc')
            micalg = self.msg.get_param('micalg')
            micalg += ',' + self._micalg(signature.hash_algorithm)
            self.msg.set_param('micalg', micalg)
            self.msg.get_payload(1).attach(sig_part)
        else:
            original_msg = copy.deepcopy(self.msg)
            to_sign = next(iter(self.get_payload()))
            signature = key.sign(to_sign, **kwargs)
            self.msg.set_payload([])
            self.msg.attach(original_msg)
            self.msg.set_type('multipart/signed')
            self.msg['MIME-Version'] = '1.0'
            self.msg.set_param('protocol', 'multipart/mixed')
            self.msg.set_param('micalg',
                               self._micalg(signature.hash_algorithm))
            sig_part = MIMEApplication(_data=str(signature),
                                       _subtype=MIMEWrapper._signature_subtype,
                                       _encoder=encode_7or8bit,
                                       name='signature.asc')
            sig_part.add_header('Content-Description',
                                'OpenPGP digital signature')
            sig_part.add_header('Content-Disposition', 'attachment',
                                filename='signature.asc')
            second_part = MIMEMultipart()
            second_part.attach(sig_part)
            self.msg.attach(second_part)
        self.msg.preamble = MIMEMultiSigWrapper._signature_preamble
        return self

    def verify(self, key):
        """
        Verify the signatures of this message with key.

        :param key: The key to verify with.
        :type key: pgpy.PGPKey
        :return: The verified signature.
        :rtype: Generator[pgpy.types.SignatureVerification]
        """
        clear_text = next(iter(self.get_signed()))
        for signature in self.get_signature():
            try:
                verification = key.verify(clear_text, signature)
            except:
                continue
            yield verification

    def decrypt(self, key):
        """
        Decrypt this message with key.

        :param key: The key to decrypt with.
        :type key: pgpy.PGPKey
        :return:
        :rtype: MIMEMultiSigWrapper
        """
        pmsg = next(iter(self.get_encrypted()))
        decrypted = key.decrypt(pmsg)

        dmsg = decrypted.message
        if isinstance(dmsg, bytearray):
            dmsg = dmsg.decode(decrypted.charset or 'utf-8')

        out = message_from_string(dmsg, _class=Message)
        if decrypted.is_signed:
            # result should be a multisig signed thing, so [out,  ]
            self.msg.set_payload([])
            self.msg.attach(out)
            self.msg.set_type('multipart/signed')
            self.msg['MIME-Version'] = '1.0'
            self.msg.set_param('protocol', 'multipart/mixed')
            micalg = ', '.join(self._micalg(sig.hash_algorithm)
                               for sig in decrypted.signatures)
            self.msg.set_param('micalg', micalg)
            second_part = MIMEMultipart()
            for signature in decrypted.signatures:
                sig = MIMEApplication(_data=str(signature),
                                      _subtype=MIMEWrapper._signature_subtype,
                                      _encoder=encode_7or8bit,
                                      name='signature.asc')
                sig.add_header('Content-Description',
                               'OpenPGP digital signature')
                sig.add_header('Content-Disposition', 'attachment',
                               filename='signature.asc')
                second_part.attach(sig)
            self.msg.attach(second_part)
        else:
            # result should be
            self.msg.set_payload(out.get_payload())
            copy_headers(out, self.msg, True)

        return self



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the PGP enabled BulkDelivery."""
import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.mailinglist import Personalization
from mailman.testing.helpers import (
    specialized_message_from_string as mfs, subscribe)

from mailman_pgp.database import transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.mta.bulk import (PGPBulkDelivery)
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key
from mailman_pgp.utils.pgp import verifies


class BulkDeliveryTester(PGPBulkDelivery):
    """Save deliveries made by the PGPBulkDelivery class."""

    def __init__(self, max_recipients=None):
        super().__init__(max_recipients=max_recipients)
        self.deliveries = []

    def _deliver_to_recipients(self, mlist, msg, msgdata, recipients):
        self.deliveries.append((mlist, msg, msgdata, recipients))
        return []


class TestPGPBulkDelivery(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.mlist = create_list('test@example.com', style_name='pgp-default')
        self.mlist.personalize = Personalization.none

        self.list_key = load_key('ecc_p256.priv.asc')
        self.pgp_list = PGPMailingList.for_list(self.mlist)
        self.pgp_list.key = self.list_key

        # Make Anne a member of this mailing list.
        self.anne = subscribe(self.mlist, 'Anne', email='anne@example.org')
        self.anne_key = load_key('rsa_1024.priv.asc')

        self.bart = subscribe(self.mlist, 'Bart', email='bart@example.org')
        self.bart_key = load_key('ecc_secp256k1.priv.asc')

        with transaction() as t:
            self.pgp_anne = PGPAddress(self.anne.address)
            self.pgp_anne.key = self.anne_key.pubkey
            self.pgp_anne.key_confirmed = True
            t.add(self.pgp_anne)

        with transaction() as t:
            self.pgp_bart = PGPAddress(self.bart.address)
            self.pgp_bart.key = self.bart_key.pubkey
            self.pgp_bart.key_confirmed = True
            t.add(self.pgp_bart)

        # Clear out any results from the previous test.
        self.msg = mfs("""\
From: anne@example.org
To: test@example.com
Subject: test

""")

    def test_sign_encrypt(self):
        with transaction():
            self.pgp_list.sign_outgoing = True
            self.pgp_list.encrypt_outgoing = True

        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'])
        agent = BulkDeliveryTester(2)
        refused = agent.deliver(self.mlist, self.msg, msgdata)

        self.assertEqual(len(refused), 0)
        self.assertEqual(len(agent.deliveries), 1)

        out_msg = agent.deliveries[0][1]
        out_wrapped = PGPWrapper(out_msg)
        self.assertTrue(out_wrapped.is_encrypted())

        decrypted = out_wrapped.copy().decrypt(self.list_key)
        self.assertTrue(decrypted.is_signed())
        self.assertTrue(verifies(decrypted.verify(self.list_key.pubkey)))

        decrypted = out_wrapped.copy().decrypt(self.anne_key)
        self.assertTrue(decrypted.is_signed())
        self.assertTrue(verifies(decrypted.verify(self.list_key.pubkey)))

        decrypted = out_wrapped.copy().decrypt(self.bart_key)
        self.assertTrue(decrypted.is_signed())
        self.assertTrue(verifies(decrypted.verify(self.list_key.pubkey)))

    def test_encrypt(self):
        with transaction():
            self.pgp_list.sign_outgoing = False
            self.pgp_list.encrypt_outgoing = True

        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'])
        agent = BulkDeliveryTester(2)
        refused = agent.deliver(self.mlist, self.msg, msgdata)

        self.assertEqual(len(refused), 0)
        self.assertEqual(len(agent.deliveries), 1)

        out_msg = agent.deliveries[0][1]
        wrapped = PGPWrapper(out_msg)
        self.assertTrue(wrapped.is_encrypted())
        wrapped.copy().decrypt(self.list_key)
        wrapped.copy().decrypt(self.anne_key)
        wrapped.copy().decrypt(self.bart_key)

    def test_sign(self):
        with transaction():
            self.pgp_list.sign_outgoing = True
            self.pgp_list.encrypt_outgoing = False

        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'])
        agent = BulkDeliveryTester(2)
        refused = agent.deliver(self.mlist, self.msg, msgdata)

        self.assertEqual(len(refused), 0)
        self.assertEqual(len(agent.deliveries), 1)

        out_msg = agent.deliveries[0][1]
        wrapped = PGPWrapper(out_msg)
        self.assertTrue(wrapped.is_signed())
        self.assertTrue(verifies(wrapped.verify(self.list_key.pubkey)))

    def test_none(self):
        with transaction():
            self.pgp_list.sign_outgoing = False
            self.pgp_list.encrypt_outgoing = False

        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'])
        agent = BulkDeliveryTester(2)
        refused = agent.deliver(self.mlist, self.msg, msgdata)

        self.assertEqual(len(refused), 0)
        self.assertEqual(len(agent.deliveries), 1)

        out_msg = agent.deliveries[0][1]
        wrapped = PGPWrapper(out_msg)
        self.assertFalse(wrapped.is_signed())
        self.assertFalse(wrapped.is_encrypted())

    def test_no_pgp_list(self):
        ordinary_list = create_list('ordinary@example.com')
        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'])
        agent = BulkDeliveryTester(2)
        refused = agent.deliver(ordinary_list, self.msg, msgdata)

        self.assertEqual(len(refused), 0)
        self.assertEqual(len(agent.deliveries), 1)

    def test_no_pgp_address(self):
        msgdata = dict(recipients=['anne@example.org', 'someone@example.org'])
        agent = BulkDeliveryTester(2)
        refused = agent.deliver(self.mlist, self.msg, msgdata)

        self.assertEqual(len(refused), 1)
        self.assertEqual(len(agent.deliveries), 1)

    def test_no_key(self):
        with transaction():
            self.pgp_bart.key = None
        msgdata = dict(recipients=['anne@example.org', 'bart@example.org'])
        agent = BulkDeliveryTester(2)
        refused = agent.deliver(self.mlist, self.msg, msgdata)

        self.assertEqual(len(refused), 1)
        self.assertEqual(len(agent.deliveries), 1)
